@@ -1,8 +1,8 @@
 /*
    +----------------------------------------------------------------------+
-   | PHP version 4.0                                                      |
+   | PHP Version 4                                                        |
    +----------------------------------------------------------------------+
-   | Copyright (c) 1997, 1998, 1999, 2000 The PHP Group                   |
+   | Copyright (c) 1997-2003 The PHP Group                                |
    +----------------------------------------------------------------------+
    | This source file is subject to version 2.02 of the PHP license,      |
    | that is bundled with this package in the file LICENSE, and is        |
@@ -17,7 +17,7 @@
    +----------------------------------------------------------------------+
 */
 
-/* $Id: tsrm_virtual_cwd.c,v 1.13 2001/01/21 17:35:31 andi Exp $ */
+/* $Id: tsrm_virtual_cwd.c,v 1.41.2.4 2003/07/28 18:35:34 iliaa Exp $ */
 
 #include <sys/types.h>
 #include <sys/stat.h>
@@ -33,16 +33,24 @@
 
 #ifdef TSRM_WIN32
 #include <io.h>
+#include "tsrm_win32.h"
+#endif
+
+#ifdef NETWARE
+/*#include "pipe.h"*/
+#include "tsrm_nw.h"
+#endif
+
+#ifndef HAVE_REALPATH
+#define realpath(x,y) strcpy(y,x)
 #endif
 
 #define VIRTUAL_CWD_DEBUG 0
 
-#ifdef ZTS
 #include "TSRM.h"
-#endif
 
-/* Only need mutex for popen() in Windows because it doesn't chdir() on UNIX */
-#if defined(TSRM_WIN32) && defined(ZTS)
+/* Only need mutex for popen() in Windows and NetWare, because it doesn't chdir() on UNIX */
+#if (defined(TSRM_WIN32) || defined(NETWARE)) && defined(ZTS)
 MUTEX_T cwd_mutex;
 #endif
 
@@ -69,7 +77,8 @@ cwd_state main_cwd_state; /* True global */
 #endif
 
 #ifdef TSRM_WIN32
-#define tsrm_strtok_r(a,b,c) strtok((a),(b))
+#include <tchar.h>
+#define tsrm_strtok_r(a,b,c) _tcstok((a),(b))
 #define TOKENIZER_STRING "/\\"
 	
 static int php_check_dots(const char *element, int n) 
@@ -83,8 +92,15 @@ static int php_check_dots(const char *element, int n)
 	(len >= 2 && !php_check_dots(element, len))
 
 #define IS_DIRECTORY_CURRENT(element, len) \
-	(len == 1 && ptr[0] == '.')
+	(len == 1 && element[0] == '.')
 
+#elif defined(NETWARE)
+/* NetWare has strtok() (in LibC) and allows both slashes in paths, like Windows --
+   but rest of the stuff is like Unix */
+/* strtok() call in LibC is abending when used in a different address space -- hence using
+   PHP's version itself for now */
+/*#define tsrm_strtok_r(a,b,c) strtok((a),(b))*/
+#define TOKENIZER_STRING "/\\"
 
 #else
 #define TOKENIZER_STRING "/"
@@ -95,12 +111,12 @@ static int php_check_dots(const char *element, int n)
 
 #ifndef IS_DIRECTORY_UP
 #define IS_DIRECTORY_UP(element, len) \
-	(len == 2 && memcmp(element, "..", 2) == 0)
+	(len == 2 && element[0] == '.' && element[1] == '.')
 #endif
 
 #ifndef IS_DIRECTORY_CURRENT
 #define IS_DIRECTORY_CURRENT(element, len) \
-	(len == 1 && ptr[0] == '.')
+	(len == 1 && element[0] == '.')
 #endif
 
 /* define this to check semantics */
@@ -121,9 +137,15 @@ static int php_check_dots(const char *element, int n)
 	
 static int php_is_dir_ok(const cwd_state *state) 
 {
+#if !(defined(NETWARE) && defined(CLIB_STAT_PATCH))
 	struct stat buf;
 
 	if (stat(state->cwd, &buf) == 0 && S_ISDIR(buf.st_mode))
+#else
+    struct stat_libc buf;
+
+    if (stat(state->cwd, (struct stat*)(&buf)) == 0 && S_ISDIR(buf.st_mode))
+#endif
 		return (0);
 
 	return (1);
@@ -131,20 +153,26 @@ static int php_is_dir_ok(const cwd_state *state)
 
 static int php_is_file_ok(const cwd_state *state) 
 {
+#if !(defined(NETWARE) && defined(CLIB_STAT_PATCH))
 	struct stat buf;
 
 	if (stat(state->cwd, &buf) == 0 && S_ISREG(buf.st_mode))
+#else
+    struct stat_libc buf;
+
+    if (stat(state->cwd, (struct stat*)(&buf)) == 0 && S_ISREG(buf.st_mode))
+#endif
 		return (0);
 
 	return (1);
 }
 
-static void cwd_globals_ctor(virtual_cwd_globals *cwd_globals)
+static void cwd_globals_ctor(virtual_cwd_globals *cwd_globals TSRMLS_DC)
 {
 	CWD_STATE_COPY(&cwd_globals->cwd, &main_cwd_state);
 }
 
-static void cwd_globals_dtor(virtual_cwd_globals *cwd_globals)
+static void cwd_globals_dtor(virtual_cwd_globals *cwd_globals TSRMLS_DC)
 {
 	CWD_STATE_FREE(&cwd_globals->cwd);
 }
@@ -177,9 +205,9 @@ CWD_API void virtual_cwd_startup(void)
 	main_cwd_state.cwd_length = strlen(cwd);
 
 #ifdef ZTS
-	cwd_globals_id = ts_allocate_id(sizeof(virtual_cwd_globals), (ts_allocate_ctor) cwd_globals_ctor, (ts_allocate_dtor) cwd_globals_dtor);
+	ts_allocate_id(&cwd_globals_id, sizeof(virtual_cwd_globals), (ts_allocate_ctor) cwd_globals_ctor, (ts_allocate_dtor) cwd_globals_dtor);
 #else
-	cwd_globals_ctor(&cwd_globals);
+	cwd_globals_ctor(&cwd_globals TSRMLS_CC);
 #endif
 
 #if defined(TSRM_WIN32) && defined(ZTS)
@@ -190,7 +218,7 @@ CWD_API void virtual_cwd_startup(void)
 CWD_API void virtual_cwd_shutdown(void)
 {
 #ifndef ZTS
-	cwd_globals_dtor(&cwd_globals);
+	cwd_globals_dtor(&cwd_globals TSRMLS_CC);
 #endif
 #if defined(TSRM_WIN32) && defined(ZTS)
 	tsrm_mutex_free(cwd_mutex);
@@ -199,10 +227,9 @@ CWD_API void virtual_cwd_shutdown(void)
 	free(main_cwd_state.cwd); /* Don't use CWD_STATE_FREE because the non global states will probably use emalloc()/efree() */
 }
 
-CWD_API char *virtual_getcwd_ex(size_t *length)
+CWD_API char *virtual_getcwd_ex(size_t *length TSRMLS_DC)
 {
 	cwd_state *state;
-	CWDLS_FETCH();
 
 	state = &CWDG(cwd);
 
@@ -235,12 +262,12 @@ CWD_API char *virtual_getcwd_ex(size_t *length)
 
 
 /* Same semantics as UNIX getcwd() */
-CWD_API char *virtual_getcwd(char *buf, size_t size)
+CWD_API char *virtual_getcwd(char *buf, size_t size TSRMLS_DC)
 {
 	size_t length;
 	char *cwd;
 
-	cwd = virtual_getcwd_ex(&length);
+	cwd = virtual_getcwd_ex(&length TSRMLS_CC);
 
 	if (buf == NULL) {
 		return cwd;
@@ -257,7 +284,7 @@ CWD_API char *virtual_getcwd(char *buf, size_t size)
 
 /* Resolve path relatively to state and put the real path into state */
 /* returns 0 for ok, 1 for error */
-CWD_API int virtual_file_ex(cwd_state *state, const char *path, verify_path_func verify_path)
+CWD_API int virtual_file_ex(cwd_state *state, const char *path, verify_path_func verify_path, int use_realpath)
 {
 	int path_length = strlen(path);
 	char *ptr, *path_copy;
@@ -275,9 +302,12 @@ CWD_API int virtual_file_ex(cwd_state *state, const char *path, verify_path_func
 	if (path_length == 0) 
 		return (0);
 
-#if !defined(TSRM_WIN32) && !defined(__BEOS__)
-	if (IS_ABSOLUTE_PATH(path, path_length)) {
-		if (realpath(path, resolved_path)) {
+#if !defined(TSRM_WIN32) && !defined(NETWARE)
+	/* cwd_length can be 0 when getcwd() fails.
+	 * This can happen under solaris when a dir does not have read permissions
+	 * but *does* have execute permissions */
+	if (IS_ABSOLUTE_PATH(path, path_length) || (state->cwd_length < 1)) {
+		if (use_realpath && realpath(path, resolved_path)) {
 			path = resolved_path;
 			path_length = strlen(path);
 		}
@@ -295,7 +325,7 @@ CWD_API int virtual_file_ex(cwd_state *state, const char *path, verify_path_func
 		memcpy(ptr, path, path_length);
 		ptr += path_length;
 		*ptr = '\0';
-		if (realpath(tmp, resolved_path)) {
+		if (use_realpath && realpath(tmp, resolved_path)) {
 			path = resolved_path;
 			path_length = strlen(path);
 		}
@@ -310,10 +340,13 @@ CWD_API int virtual_file_ex(cwd_state *state, const char *path, verify_path_func
 	fprintf(stderr,"cwd = %s path = %s\n", state->cwd, path);
 #endif
 	if (IS_ABSOLUTE_PATH(path_copy, path_length)) {
-		copy_amount = COPY_WHEN_ABSOLUTE;
+        copy_amount = COPY_WHEN_ABSOLUTE(path_copy);
 		is_absolute = 1;
 #ifdef TSRM_WIN32
-	} else if(IS_SLASH(path_copy[0])) {
+	} else if (IS_UNC_PATH(path_copy, path_length)) {
+		copy_amount = 2;
+		is_absolute = 1;
+	} else if (IS_SLASH(path_copy[0])) {
 		copy_amount = 2;
 #endif
 	}
@@ -333,45 +366,64 @@ CWD_API int virtual_file_ex(cwd_state *state, const char *path, verify_path_func
 	}
 
 
-	ptr = tsrm_strtok_r(path_copy, TOKENIZER_STRING, &tok);
-	while (ptr) {
-		ptr_length = strlen(ptr);
+	if (state->cwd_length > 0 || IS_ABSOLUTE_PATH(path, path_length)) {
+		ptr = tsrm_strtok_r(path_copy, TOKENIZER_STRING, &tok);
+		while (ptr) {
+			ptr_length = strlen(ptr);
 
-		if (IS_DIRECTORY_UP(ptr, ptr_length)) {
-			char save;
+			if (IS_DIRECTORY_UP(ptr, ptr_length)) {
+				char save;
 
-			save = DEFAULT_SLASH;
+				save = DEFAULT_SLASH;
 
 #define PREVIOUS state->cwd[state->cwd_length - 1]
 
-			while (IS_ABSOLUTE_PATH(state->cwd, state->cwd_length) &&
-					!IS_SLASH(PREVIOUS)) {
-				save = PREVIOUS;
-				PREVIOUS = '\0';
-				state->cwd_length--;
-			}
+				while (IS_ABSOLUTE_PATH(state->cwd, state->cwd_length) &&
+						!IS_SLASH(PREVIOUS)) {
+					save = PREVIOUS;
+					PREVIOUS = '\0';
+					state->cwd_length--;
+				}
 
-			if (!IS_ABSOLUTE_PATH(state->cwd, state->cwd_length)) {
-				state->cwd[state->cwd_length++] = save;
-				state->cwd[state->cwd_length] = '\0';
-			} else {
-				PREVIOUS = '\0';
-				state->cwd_length--;
+				if (!IS_ABSOLUTE_PATH(state->cwd, state->cwd_length)) {
+					state->cwd[state->cwd_length++] = save;
+					state->cwd[state->cwd_length] = '\0';
+				} else {
+					PREVIOUS = '\0';
+					state->cwd_length--;
+				}
+			} else if (!IS_DIRECTORY_CURRENT(ptr, ptr_length)) {
+				state->cwd = (char *) realloc(state->cwd, state->cwd_length+ptr_length+1+1);
+#ifdef TSRM_WIN32
+				/* Windows 9x will consider C:\\Foo as a network path. Avoid it. */
+				if ((state->cwd[state->cwd_length-1]!='\\' && state->cwd[state->cwd_length-1]!='/') || 
+						IsDBCSLeadByte(state->cwd[state->cwd_length-2])) {
+					state->cwd[state->cwd_length++] = DEFAULT_SLASH;
+				}
+#elif defined(NETWARE)
+				/* If the token is a volume name, it will have colon at the end -- so, no slash before it */
+				if (ptr[ptr_length-1] != ':') {
+					state->cwd[state->cwd_length++] = DEFAULT_SLASH;
+				}
+#else
+				state->cwd[state->cwd_length++] = DEFAULT_SLASH;
+#endif
+				memcpy(&state->cwd[state->cwd_length], ptr, ptr_length+1);
+				state->cwd_length += ptr_length;
 			}
-		} else if (!IS_DIRECTORY_CURRENT(ptr, ptr_length)) {
-			state->cwd = (char *) realloc(state->cwd, state->cwd_length+ptr_length+1+1);
-			state->cwd[state->cwd_length] = DEFAULT_SLASH;
-			memcpy(&state->cwd[state->cwd_length+1], ptr, ptr_length+1);
-			state->cwd_length += (ptr_length+1);
+			ptr = tsrm_strtok_r(NULL, TOKENIZER_STRING, &tok);
 		}
-		ptr = tsrm_strtok_r(NULL, TOKENIZER_STRING, &tok);
-	}
 
-	if (state->cwd_length == COPY_WHEN_ABSOLUTE) {
-		state->cwd = (char *) realloc(state->cwd, state->cwd_length+1+1);
-		state->cwd[state->cwd_length] = DEFAULT_SLASH;
-		state->cwd[state->cwd_length+1] = '\0';
-		state->cwd_length++;
+		if (state->cwd_length == COPY_WHEN_ABSOLUTE(state->cwd)) {
+			state->cwd = (char *) realloc(state->cwd, state->cwd_length+1+1);
+			state->cwd[state->cwd_length] = DEFAULT_SLASH;
+			state->cwd[state->cwd_length+1] = '\0';
+			state->cwd_length++;
+		}
+	} else {
+		state->cwd = (char *) realloc(state->cwd, path_length+1);
+		memcpy(state->cwd, path, path_length+1);
+		state->cwd_length = path_length;
 	}
 
 	if (verify_path && verify_path(state)) {
@@ -394,14 +446,12 @@ CWD_API int virtual_file_ex(cwd_state *state, const char *path, verify_path_func
 	return (ret);
 }
 
-CWD_API int virtual_chdir(const char *path)
+CWD_API int virtual_chdir(const char *path TSRMLS_DC)
 {
-	CWDLS_FETCH();
-
-	return virtual_file_ex(&CWDG(cwd), path, php_is_dir_ok)?-1:0;
+	return virtual_file_ex(&CWDG(cwd), path, php_is_dir_ok, 1)?-1:0;
 }
 
-CWD_API int virtual_chdir_file(const char *path, int (*p_chdir)(const char *path))
+CWD_API int virtual_chdir_file(const char *path, int (*p_chdir)(const char *path TSRMLS_DC) TSRMLS_DC)
 {
 	int length = strlen(path);
 	char *temp;
@@ -419,7 +469,7 @@ CWD_API int virtual_chdir_file(const char *path, int (*p_chdir)(const char *path
 		return -1;
 	}
 
-	if (length == COPY_WHEN_ABSOLUTE && IS_ABSOLUTE_PATH(path, length+1)) { /* Also use trailing slash if this is absolute */
+	if (length == COPY_WHEN_ABSOLUTE(path) && IS_ABSOLUTE_PATH(path, length+1)) { /* Also use trailing slash if this is absolute */
 		length++;
 	}
 	temp = (char *) tsrm_do_alloca(length+1);
@@ -428,38 +478,40 @@ CWD_API int virtual_chdir_file(const char *path, int (*p_chdir)(const char *path
 #if VIRTUAL_CWD_DEBUG
 	fprintf (stderr, "Changing directory to %s\n", temp);
 #endif
-	retval = p_chdir(temp);
+	retval = p_chdir(temp TSRMLS_CC);
 	tsrm_free_alloca(temp);
 	return retval;
 }
 
-CWD_API char *virtual_realpath(const char *path, char *real_path)
+CWD_API char *virtual_realpath(const char *path, char *real_path TSRMLS_DC)
 {
 	cwd_state new_state;
-	int retval;
-    CWDLS_FETCH();
+	char *retval;
 
 	CWD_STATE_COPY(&new_state, &CWDG(cwd));
-	retval = virtual_file_ex(&new_state, path, NULL);
 	
-	if (!retval) {
+	if (virtual_file_ex(&new_state, path, NULL, 1)==0) {
 		int len = new_state.cwd_length>MAXPATHLEN-1?MAXPATHLEN-1:new_state.cwd_length;
+
 		memcpy(real_path, new_state.cwd, len);
 		real_path[len] = '\0';
-		return real_path;
+		retval = real_path;
+	} else {
+		retval = NULL;
 	}
 
-	return NULL;
+	CWD_STATE_FREE(&new_state);
+
+	return retval;
 }
 
-CWD_API int virtual_filepath_ex(const char *path, char **filepath, verify_path_func verify_path)
+CWD_API int virtual_filepath_ex(const char *path, char **filepath, verify_path_func verify_path TSRMLS_DC)
 {
 	cwd_state new_state;
 	int retval;
-	CWDLS_FETCH();
 
 	CWD_STATE_COPY(&new_state, &CWDG(cwd));
-	retval = virtual_file_ex(&new_state, path, verify_path);
+	retval = virtual_file_ex(&new_state, path, verify_path, 1);
 
 	*filepath = new_state.cwd;
 
@@ -467,23 +519,22 @@ CWD_API int virtual_filepath_ex(const char *path, char **filepath, verify_path_f
 
 }
 
-CWD_API int virtual_filepath(const char *path, char **filepath)
+CWD_API int virtual_filepath(const char *path, char **filepath TSRMLS_DC)
 {
-	return virtual_filepath_ex(path, filepath, php_is_file_ok);
+	return virtual_filepath_ex(path, filepath, php_is_file_ok TSRMLS_CC);
 }
 
-CWD_API FILE *virtual_fopen(const char *path, const char *mode)
+CWD_API FILE *virtual_fopen(const char *path, const char *mode TSRMLS_DC)
 {
 	cwd_state new_state;
 	FILE *f;
-	CWDLS_FETCH();
 
 	if (path[0] == '\0') { /* Fail to open empty path */
 		return NULL;
 	}
 
 	CWD_STATE_COPY(&new_state, &CWDG(cwd));
-	virtual_file_ex(&new_state, path, NULL);
+	virtual_file_ex(&new_state, path, NULL, 1);
 
 	f = fopen(new_state.cwd, mode);
 
@@ -491,15 +542,32 @@ CWD_API FILE *virtual_fopen(const char *path, const char *mode)
 	return f;
 }
 
+#if !defined(TSRM_WIN32)
+CWD_API int virtual_access(const char *pathname, int mode TSRMLS_DC)
+{
+	cwd_state new_state;
+	int ret;
+	
+	CWD_STATE_COPY(&new_state, &CWDG(cwd));
+	virtual_file_ex(&new_state, pathname, NULL, 1);
+
+	ret = access(new_state.cwd, mode);
+	
+	CWD_STATE_FREE(&new_state);
+	
+	return ret;
+}
+#endif
+
+
 #if HAVE_UTIME
-CWD_API int virtual_utime(const char *filename, struct utimbuf *buf)
+CWD_API int virtual_utime(const char *filename, struct utimbuf *buf TSRMLS_DC)
 {
 	cwd_state new_state;
 	int ret;
-	CWDLS_FETCH();
 
 	CWD_STATE_COPY(&new_state, &CWDG(cwd));
-	virtual_file_ex(&new_state, filename, NULL);
+	virtual_file_ex(&new_state, filename, NULL, 0);
 
 	ret = utime(new_state.cwd, buf);
 
@@ -508,14 +576,13 @@ CWD_API int virtual_utime(const char *filename, struct utimbuf *buf)
 }
 #endif
 
-CWD_API int virtual_chmod(const char *filename, mode_t mode)
+CWD_API int virtual_chmod(const char *filename, mode_t mode TSRMLS_DC)
 {
 	cwd_state new_state;
 	int ret;
-	CWDLS_FETCH();
 
 	CWD_STATE_COPY(&new_state, &CWDG(cwd));
-	virtual_file_ex(&new_state, filename, NULL);
+	virtual_file_ex(&new_state, filename, NULL, 1);
 
 	ret = chmod(new_state.cwd, mode);
 
@@ -523,15 +590,14 @@ CWD_API int virtual_chmod(const char *filename, mode_t mode)
 	return ret;
 }
 
-#ifndef TSRM_WIN32
-CWD_API int virtual_chown(const char *filename, uid_t owner, gid_t group)
+#if !defined(TSRM_WIN32) && !defined(NETWARE)
+CWD_API int virtual_chown(const char *filename, uid_t owner, gid_t group TSRMLS_DC)
 {
 	cwd_state new_state;
 	int ret;
-	CWDLS_FETCH();
 
 	CWD_STATE_COPY(&new_state, &CWDG(cwd));
-	virtual_file_ex(&new_state, filename, NULL);
+	virtual_file_ex(&new_state, filename, NULL, 0);
 
 	ret = chown(new_state.cwd, owner, group);
 
@@ -540,21 +606,20 @@ CWD_API int virtual_chown(const char *filename, uid_t owner, gid_t group)
 }
 #endif
 
-CWD_API int virtual_open(const char *path, int flags, ...)
+CWD_API int virtual_open(const char *path TSRMLS_DC, int flags, ...)
 {
 	cwd_state new_state;
 	int f;
-	CWDLS_FETCH();
 
 	CWD_STATE_COPY(&new_state, &CWDG(cwd));
-	virtual_file_ex(&new_state, path, NULL);
+	virtual_file_ex(&new_state, path, NULL, 1);
 
 	if (flags & O_CREAT) {
 		mode_t mode;
 		va_list arg;
 
 		va_start(arg, flags);
-		mode = va_arg(arg, mode_t);
+		mode = (mode_t) va_arg(arg, int);
 		va_end(arg);
 
 		f = open(new_state.cwd, flags, mode);
@@ -565,14 +630,13 @@ CWD_API int virtual_open(const char *path, int flags, ...)
 	return f;
 }
 
-CWD_API int virtual_creat(const char *path, mode_t mode)
+CWD_API int virtual_creat(const char *path, mode_t mode TSRMLS_DC)
 {
 	cwd_state new_state;
 	int f;
-	CWDLS_FETCH();
 
 	CWD_STATE_COPY(&new_state, &CWDG(cwd));
-	virtual_file_ex(&new_state, path, NULL);
+	virtual_file_ex(&new_state, path, NULL, 1);
 
 	f = creat(new_state.cwd,  mode);
 
@@ -580,19 +644,18 @@ CWD_API int virtual_creat(const char *path, mode_t mode)
 	return f;
 }
 
-CWD_API int virtual_rename(char *oldname, char *newname)
+CWD_API int virtual_rename(char *oldname, char *newname TSRMLS_DC)
 {
 	cwd_state old_state;
 	cwd_state new_state;
 	int retval;
-	CWDLS_FETCH();
 
 	CWD_STATE_COPY(&old_state, &CWDG(cwd));
-	virtual_file_ex(&old_state, oldname, NULL);
+	virtual_file_ex(&old_state, oldname, NULL, 0);
 	oldname = old_state.cwd;
 
 	CWD_STATE_COPY(&new_state, &CWDG(cwd));
-	virtual_file_ex(&new_state, newname, NULL);
+	virtual_file_ex(&new_state, newname, NULL, 0);
 	newname = new_state.cwd;
  
 	retval = rename(oldname, newname);
@@ -603,48 +666,59 @@ CWD_API int virtual_rename(char *oldname, char *newname)
 	return retval;
 }
 
-CWD_API int virtual_stat(const char *path, struct stat *buf)
+#if !(defined(NETWARE) && defined(CLIB_STAT_PATCH))
+CWD_API int virtual_stat(const char *path, struct stat *buf TSRMLS_DC)
 {
 	cwd_state new_state;
 	int retval;
-	CWDLS_FETCH();
 
 	CWD_STATE_COPY(&new_state, &CWDG(cwd));
-	virtual_file_ex(&new_state, path, NULL);
+	virtual_file_ex(&new_state, path, NULL, 1);
 
 	retval = stat(new_state.cwd, buf);
 
 	CWD_STATE_FREE(&new_state);
 	return retval;
 }
-
-#ifndef TSRM_WIN32
-
-CWD_API int virtual_lstat(const char *path, struct stat *buf)
+#else
+CWD_API int virtual_stat(const char *path, struct stat_libc *buf TSRMLS_DC)
 {
 	cwd_state new_state;
 	int retval;
-	CWDLS_FETCH();
 
 	CWD_STATE_COPY(&new_state, &CWDG(cwd));
-	virtual_file_ex(&new_state, path, NULL);
+	virtual_file_ex(&new_state, path, NULL, 1);
+
+	retval = stat(new_state.cwd, (struct stat*)buf);
+
+	CWD_STATE_FREE(&new_state);
+	return retval;
+}
+#endif
+
+#if !defined(TSRM_WIN32) && !defined(NETWARE)
+CWD_API int virtual_lstat(const char *path, struct stat *buf TSRMLS_DC)
+{
+	cwd_state new_state;
+	int retval;
+
+	CWD_STATE_COPY(&new_state, &CWDG(cwd));
+	virtual_file_ex(&new_state, path, NULL, 0);
 
 	retval = lstat(new_state.cwd, buf);
 
 	CWD_STATE_FREE(&new_state);
 	return retval;
 }
-
 #endif
 
-CWD_API int virtual_unlink(const char *path)
+CWD_API int virtual_unlink(const char *path TSRMLS_DC)
 {
 	cwd_state new_state;
 	int retval;
-	CWDLS_FETCH();
 
 	CWD_STATE_COPY(&new_state, &CWDG(cwd));
-	virtual_file_ex(&new_state, path, NULL);
+	virtual_file_ex(&new_state, path, NULL, 0);
 
 	retval = unlink(new_state.cwd);
 
@@ -652,14 +726,13 @@ CWD_API int virtual_unlink(const char *path)
 	return retval;
 }
 
-CWD_API int virtual_mkdir(const char *pathname, mode_t mode)
+CWD_API int virtual_mkdir(const char *pathname, mode_t mode TSRMLS_DC)
 {
 	cwd_state new_state;
 	int retval;
-	CWDLS_FETCH();
 
 	CWD_STATE_COPY(&new_state, &CWDG(cwd));
-	virtual_file_ex(&new_state, pathname, NULL);
+	virtual_file_ex(&new_state, pathname, NULL, 1);
 
 #ifdef TSRM_WIN32
 	retval = mkdir(new_state.cwd);
@@ -670,14 +743,13 @@ CWD_API int virtual_mkdir(const char *pathname, mode_t mode)
 	return retval;
 }
 
-CWD_API int virtual_rmdir(const char *pathname)
+CWD_API int virtual_rmdir(const char *pathname TSRMLS_DC)
 {
 	cwd_state new_state;
 	int retval;
-	CWDLS_FETCH();
 
 	CWD_STATE_COPY(&new_state, &CWDG(cwd));
-	virtual_file_ex(&new_state, pathname, NULL);
+	virtual_file_ex(&new_state, pathname, NULL, 0);
 
 	retval = rmdir(new_state.cwd);
 
@@ -689,14 +761,13 @@ CWD_API int virtual_rmdir(const char *pathname)
 DIR *opendir(const char *name);
 #endif
 
-CWD_API DIR *virtual_opendir(const char *pathname)
+CWD_API DIR *virtual_opendir(const char *pathname TSRMLS_DC)
 {
 	cwd_state new_state;
 	DIR *retval;
-	CWDLS_FETCH();
 
 	CWD_STATE_COPY(&new_state, &CWDG(cwd));
-	virtual_file_ex(&new_state, pathname, NULL);
+	virtual_file_ex(&new_state, pathname, NULL, 1);
 
 	retval = opendir(new_state.cwd);
 
@@ -704,15 +775,52 @@ CWD_API DIR *virtual_opendir(const char *pathname)
 	return retval;
 }
 
-#ifndef TSRM_WIN32
+#ifdef TSRM_WIN32
 
-CWD_API FILE *virtual_popen(const char *command, const char *type)
+CWD_API FILE *virtual_popen(const char *command, const char *type TSRMLS_DC)
+{
+	return popen_ex(command, type, CWDG(cwd).cwd, NULL);
+}
+
+#elif defined(NETWARE)
+
+/* On NetWare, the trick of prepending "cd cwd; " doesn't work so we need to perform
+   a VCWD_CHDIR() and mutex it
+ */
+CWD_API FILE *virtual_popen(const char *command, const char *type TSRMLS_DC)
+{
+	char prev_cwd[MAXPATHLEN];
+	char *getcwd_result;
+	FILE *retval;
+
+	getcwd_result = VCWD_GETCWD(prev_cwd, MAXPATHLEN);
+	if (!getcwd_result) {
+		return NULL;
+	}
+
+#ifdef ZTS
+	tsrm_mutex_lock(cwd_mutex);
+#endif
+
+	VCWD_CHDIR(CWDG(cwd).cwd);
+	retval = popen(command, type);
+	VCWD_CHDIR(prev_cwd);
+
+#ifdef ZTS
+	tsrm_mutex_unlock(cwd_mutex);
+#endif
+
+	return retval;
+}
+
+#else /* Unix */
+
+CWD_API FILE *virtual_popen(const char *command, const char *type TSRMLS_DC)
 {
 	int command_length;
 	char *command_line;
 	char *ptr;
 	FILE *retval;
-	CWDLS_FETCH();
 
 	command_length = strlen(command);
 
@@ -735,47 +843,12 @@ CWD_API FILE *virtual_popen(const char *command, const char *type)
 	*ptr++ = ' ';
 
 	memcpy(ptr, command, command_length+1);
-#ifdef TSRM_WIN32
-	retval = _popen(command_line, type);
-#else
 	retval = popen(command_line, type);
-#endif
+
 	free(command_line);
 	return retval;
 }
 
-#else
-
-/* On Windows the trick of prepending "cd cwd; " doesn't work so we need to perform
-   a real chdir() and mutex it
- */
-CWD_API FILE *virtual_popen(const char *command, const char *type)
-{
-	char prev_cwd[MAXPATHLEN];
-	char *getcwd_result;
-	FILE *retval;
-	CWDLS_FETCH();
-
-	getcwd_result = getcwd(prev_cwd, MAXPATHLEN);
-	if (!getcwd_result) {
-		return NULL;
-	}
-
-#ifdef ZTS
-	tsrm_mutex_lock(cwd_mutex);
-#endif
-
-	chdir(CWDG(cwd).cwd);
-	retval = _popen(command, type);
-	chdir(prev_cwd);
-
-#ifdef ZTS
-	tsrm_mutex_unlock(cwd_mutex);
-#endif
-
-	return retval;
-}
-
 #endif
 
 
