@@ -2,12 +2,12 @@
    +----------------------------------------------------------------------+
    | Zend Engine                                                          |
    +----------------------------------------------------------------------+
-   | Copyright (c) 1998-2001 Zend Technologies Ltd. (http://www.zend.com) |
+   | Copyright (c) 1998-2003 Zend Technologies Ltd. (http://www.zend.com) |
    +----------------------------------------------------------------------+
-   | This source file is subject to version 0.92 of the Zend license,     |
+   | This source file is subject to version 2.00 of the Zend license,     |
    | that is bundled with this package in the file LICENSE, and is        |
    | available at through the world-wide-web at                           |
-   | http://www.zend.com/license/0_92.txt.                                |
+   | http://www.zend.com/license/2_00.txt.                                |
    | If you did not receive a copy of the Zend license and are unable to  |
    | obtain it through the world-wide-web, please send a note to          |
    | license@zend.com so we can mail you a copy immediately.              |
@@ -22,7 +22,7 @@
 #include "zend_API.h"
 #include "zend_builtin_functions.h"
 #include "zend_constants.h"
-
+#include "zend_ini.h"
 #undef ZEND_TEST_EXCEPTIONS
 
 static ZEND_FUNCTION(zend_version);
@@ -43,12 +43,15 @@ static ZEND_FUNCTION(get_parent_class);
 static ZEND_FUNCTION(method_exists);
 static ZEND_FUNCTION(class_exists);
 static ZEND_FUNCTION(function_exists);
+#if ZEND_DEBUG
 static ZEND_FUNCTION(leak);
 #ifdef ZEND_TEST_EXCEPTIONS
 static ZEND_FUNCTION(crash);
 #endif
+#endif
 static ZEND_FUNCTION(get_included_files);
 static ZEND_FUNCTION(is_subclass_of);
+static ZEND_FUNCTION(is_a);
 static ZEND_FUNCTION(get_class_vars);
 static ZEND_FUNCTION(get_object_vars);
 static ZEND_FUNCTION(get_class_methods);
@@ -60,14 +63,18 @@ static ZEND_FUNCTION(get_defined_functions);
 static ZEND_FUNCTION(get_defined_vars);
 static ZEND_FUNCTION(create_function);
 static ZEND_FUNCTION(get_resource_type);
+static ZEND_FUNCTION(get_loaded_extensions);
+static ZEND_FUNCTION(extension_loaded);
+static ZEND_FUNCTION(get_extension_funcs);
+static ZEND_FUNCTION(get_defined_constants);
+static ZEND_FUNCTION(debug_backtrace);
 #if ZEND_DEBUG
 static ZEND_FUNCTION(zend_test_func);
 #endif
 
 ZEND_API unsigned char first_arg_force_ref[] = { 1, BYREF_FORCE };
-ZEND_API unsigned char first_arg_allow_ref[] = { 1, BYREF_ALLOW };
 ZEND_API unsigned char second_arg_force_ref[] = { 2, BYREF_NONE, BYREF_FORCE };
-ZEND_API unsigned char second_arg_allow_ref[] = { 2, BYREF_NONE, BYREF_ALLOW };
+ZEND_API unsigned char third_arg_force_ref[] = { 3, BYREF_NONE, BYREF_NONE, BYREF_FORCE };
 
 static zend_function_entry builtin_functions[] = {
 	ZEND_FE(zend_version,		NULL)
@@ -88,13 +95,16 @@ static zend_function_entry builtin_functions[] = {
 	ZEND_FE(method_exists,		NULL)
 	ZEND_FE(class_exists,		NULL)
 	ZEND_FE(function_exists,	NULL)
+#if ZEND_DEBUG
 	ZEND_FE(leak,				NULL)
 #ifdef ZEND_TEST_EXCEPTIONS
 	ZEND_FE(crash,				NULL)
+#endif
 #endif
 	ZEND_FE(get_included_files,	NULL)
 	ZEND_FALIAS(get_required_files,	get_included_files,		NULL)
 	ZEND_FE(is_subclass_of,		NULL)
+	ZEND_FE(is_a,				NULL)
 	ZEND_FE(get_class_vars,		NULL)
 	ZEND_FE(get_object_vars,	NULL)
 	ZEND_FE(get_class_methods,	NULL)
@@ -107,6 +117,11 @@ static zend_function_entry builtin_functions[] = {
 	ZEND_FE(get_defined_vars,	NULL)
 	ZEND_FE(create_function,	NULL)
 	ZEND_FE(get_resource_type,	NULL)
+	ZEND_FE(get_loaded_extensions,		NULL)
+	ZEND_FE(extension_loaded,			NULL)
+	ZEND_FE(get_extension_funcs,		NULL)
+	ZEND_FE(get_defined_constants,		NULL)
+	ZEND_FE(debug_backtrace,			NULL)
 #if ZEND_DEBUG
 	ZEND_FE(zend_test_func,		NULL)
 #endif
@@ -114,18 +129,23 @@ static zend_function_entry builtin_functions[] = {
 };
 
 
-int zend_startup_builtin_functions()
+int zend_startup_builtin_functions(TSRMLS_D)
 {
-	return zend_register_functions(builtin_functions, NULL, MODULE_PERSISTENT);
+	return zend_register_functions(builtin_functions, NULL, MODULE_PERSISTENT TSRMLS_CC);
 }
 
 
+/* {{{ proto string zend_version(void)
+   Get the version of the Zend Engine */
 ZEND_FUNCTION(zend_version)
 {
 	RETURN_STRINGL(ZEND_VERSION, sizeof(ZEND_VERSION)-1, 1);
 }
+/* }}} */
 
 
+/* {{{ proto int func_num_args(void)
+   Get the number of arguments that were passed to the function */
 ZEND_FUNCTION(func_num_args)
 {
 	void **p;
@@ -145,8 +165,11 @@ ZEND_FUNCTION(func_num_args)
 		RETURN_LONG(-1);
 	}
 }
+/* }}} */
 
 
+/* {{{ proto mixed func_get_arg(int arg_num)
+   Get the $arg_num'th argument that was passed to the function */
 ZEND_FUNCTION(func_get_arg)
 {
 	void **p;
@@ -161,6 +184,11 @@ ZEND_FUNCTION(func_get_arg)
 	convert_to_long_ex(z_requested_offset);
 	requested_offset = (*z_requested_offset)->value.lval;
 
+	if (requested_offset < 0) {
+		zend_error(E_WARNING, "func_get_arg():  The argument number should be >= 0");
+		RETURN_FALSE;
+	}
+
 	p = EG(argument_stack).top_element-1-1;
 	arg_count = (ulong) *p;		/* this is the amount of arguments passed to func_get_arg(); */
 	p -= 1+arg_count;
@@ -175,7 +203,7 @@ ZEND_FUNCTION(func_get_arg)
 	arg_count = (ulong) *p;
 
 	if (requested_offset>=arg_count) {
-		zend_error(E_WARNING, "func_get_arg():  Argument %d not passed to function", requested_offset);
+		zend_error(E_WARNING, "func_get_arg():  Argument %ld not passed to function", requested_offset);
 		RETURN_FALSE;
 	}
 
@@ -183,8 +211,11 @@ ZEND_FUNCTION(func_get_arg)
 	*return_value = *arg;
 	zval_copy_ctor(return_value);
 }
+/* }}} */
 
 
+/* {{{ proto array func_get_args()
+   Get an array of the arguments that were passed to the function */
 ZEND_FUNCTION(func_get_args)
 {
 	void **p;
@@ -217,6 +248,7 @@ ZEND_FUNCTION(func_get_args)
 		zend_hash_next_index_insert(return_value->value.ht, &element, sizeof(zval *), NULL);
 	}
 }
+/* }}} */
 
 
 /* {{{ proto int strlen(string str)
@@ -233,6 +265,7 @@ ZEND_NAMED_FUNCTION(zend_if_strlen)
 }
 /* }}} */
 
+
 /* {{{ proto int strcmp(string str1, string str2)
    Binary safe string comparison */
 ZEND_FUNCTION(strcmp)
@@ -244,10 +277,11 @@ ZEND_FUNCTION(strcmp)
 	}
 	convert_to_string_ex(s1);
 	convert_to_string_ex(s2);
-	RETURN_LONG(zend_binary_zval_strcmp(*s1,*s2));
+	RETURN_LONG(zend_binary_zval_strcmp(*s1, *s2));
 }
 /* }}} */
 
+
 /* {{{ proto int strncmp(string str1, string str2, int len)
    Binary safe string comparison */
 ZEND_FUNCTION(strncmp)
@@ -260,10 +294,11 @@ ZEND_FUNCTION(strncmp)
 	convert_to_string_ex(s1);
 	convert_to_string_ex(s2);
 	convert_to_long_ex(s3);
-	RETURN_LONG(zend_binary_zval_strncmp(*s1,*s2,*s3));
+	RETURN_LONG(zend_binary_zval_strncmp(*s1, *s2, *s3));
 }
 /* }}} */
 
+
 /* {{{ proto int strcasecmp(string str1, string str2)
    Binary safe case-insensitive string comparison */
 ZEND_FUNCTION(strcasecmp)
@@ -279,6 +314,7 @@ ZEND_FUNCTION(strcasecmp)
 }
 /* }}} */
 
+
 /* {{{ proto int strncasecmp(string str1, string str2, int len)
    Binary safe string comparison */
 ZEND_FUNCTION(strncasecmp)
@@ -291,14 +327,18 @@ ZEND_FUNCTION(strncasecmp)
 	convert_to_string_ex(s1);
 	convert_to_string_ex(s2);
 	convert_to_long_ex(s3);
-	RETURN_LONG(zend_binary_zval_strncasecmp(*s1,*s2,*s3));
+	RETURN_LONG(zend_binary_zval_strncasecmp(*s1, *s2, *s3));
 }
 /* }}} */
 
+
+/* {{{ proto array each(array arr)
+   Return the currently pointed key..value pair in the passed array, and advance the pointer to the next element */
 ZEND_FUNCTION(each)
 {
-	zval **array,*entry,**entry_ptr, *tmp;
+	zval **array, *entry, **entry_ptr, *tmp;
 	char *string_key;
+	uint string_key_len;
 	ulong num_key;
 	zval **inserted_pointer;
 	HashTable *target_hash;
@@ -306,6 +346,7 @@ ZEND_FUNCTION(each)
 	if (ZEND_NUM_ARGS() != 1 || zend_get_parameters_ex(1, &array) == FAILURE) {
 		ZEND_WRONG_PARAM_COUNT();
 	}
+
 	target_hash = HASH_OF(*array);
 	if (!target_hash) {
 		zend_error(E_WARNING,"Variable passed to each() is not an array or object");
@@ -332,19 +373,23 @@ ZEND_FUNCTION(each)
 	entry->refcount++;
 
 	/* add the key elements */
-	switch (zend_hash_get_current_key(target_hash, &string_key, &num_key, 1)) {
+	switch (zend_hash_get_current_key_ex(target_hash, &string_key, &string_key_len, &num_key, 1, NULL)) {
 		case HASH_KEY_IS_STRING:
-			add_get_index_string(return_value,0,string_key,(void **) &inserted_pointer,0);
+			add_get_index_stringl(return_value, 0, string_key, string_key_len-1, (void **) &inserted_pointer, 0);
 			break;
 		case HASH_KEY_IS_LONG:
-			add_get_index_long(return_value,0,num_key, (void **) &inserted_pointer);
+			add_get_index_long(return_value, 0, num_key, (void **) &inserted_pointer);
 			break;
 	}
 	zend_hash_update(return_value->value.ht, "key", sizeof("key"), inserted_pointer, sizeof(zval *), NULL);
 	(*inserted_pointer)->refcount++;
 	zend_hash_move_forward(target_hash);
 }
+/* }}} */
 
+
+/* {{{ proto int error_reporting(int new_error_level=null)
+   Return the current error_reporting level, and if an argument was passed - change to the new level */
 ZEND_FUNCTION(error_reporting)
 {
 	zval **arg;
@@ -355,11 +400,11 @@ ZEND_FUNCTION(error_reporting)
 		case 0:
 			break;
 		case 1:
-			if (zend_get_parameters_ex(1,&arg) == FAILURE) {
+			if (zend_get_parameters_ex(1, &arg) == FAILURE) {
 				RETURN_FALSE;
 			}
-			convert_to_long_ex(arg);
-			EG(error_reporting)=(*arg)->value.lval;
+			convert_to_string_ex(arg);
+			zend_alter_ini_entry("error_reporting", sizeof("error_reporting"), Z_STRVAL_PP(arg), Z_STRLEN_PP(arg), ZEND_INI_USER, ZEND_INI_STAGE_RUNTIME);
 			break;
 		default:
 			ZEND_WRONG_PARAM_COUNT();
@@ -368,7 +413,11 @@ ZEND_FUNCTION(error_reporting)
 
 	RETVAL_LONG(old_error_reporting);
 }
+/* }}} */
 
+
+/* {{{ proto bool define(string constant_name, mixed value, case_sensitive=true)
+   Define a new constant */
 ZEND_FUNCTION(define)
 {
 	zval **var, **val, **non_cs;
@@ -418,14 +467,17 @@ ZEND_FUNCTION(define)
 	c.flags = case_sensitive; /* non persistent */
 	c.name = zend_strndup((*var)->value.str.val, (*var)->value.str.len);
 	c.name_len = (*var)->value.str.len+1;
-	if(zend_register_constant(&c ELS_CC) == SUCCESS) {
-	  RETURN_TRUE;
+	if (zend_register_constant(&c TSRMLS_CC) == SUCCESS) {
+		RETURN_TRUE;
 	} else {
-	  RETURN_FALSE;
+		RETURN_FALSE;
 	}
 }
+/* }}} */
 
 
+/* {{{ proto bool defined(string constant_name)
+   Check whether a constant exists */
 ZEND_FUNCTION(defined)
 {
 	zval **var;
@@ -436,68 +488,87 @@ ZEND_FUNCTION(defined)
 	}
 	
 	convert_to_string_ex(var);
-	if (zend_get_constant((*var)->value.str.val, (*var)->value.str.len, &c)) {
+	if (zend_get_constant((*var)->value.str.val, (*var)->value.str.len, &c TSRMLS_CC)) {
 		zval_dtor(&c);
 		RETURN_LONG(1);
 	} else {
 		RETURN_LONG(0);
 	}
 }
+/* }}} */
+
 
 /* {{{ proto string get_class(object object)
    Retrieves the class name */
 ZEND_FUNCTION(get_class)
 {
 	zval **arg;
+	zend_class_entry *ce;
 	
 	if (ZEND_NUM_ARGS()!=1 || zend_get_parameters_ex(1, &arg)==FAILURE) {
 		ZEND_WRONG_PARAM_COUNT();
 	}
-	if ((*arg)->type != IS_OBJECT) {
+	if (Z_TYPE_PP(arg) != IS_OBJECT) {
 		RETURN_FALSE;
 	}
-	RETURN_STRINGL((*arg)->value.obj.ce->name,	(*arg)->value.obj.ce->name_length, 1);
+
+	ce = Z_OBJCE_PP(arg);
+	RETURN_STRINGL(ce->name, ce->name_length, 1);
 }
 /* }}} */
 
-/* {{{ proto string get_parent_class(object object)
-   Retrieves the parent class name */
+
+/* {{{ proto string get_parent_class(mixed object)
+   Retrieves the parent class name for object or class. */
 ZEND_FUNCTION(get_parent_class)
 {
 	zval **arg;
+	zend_class_entry *ce = NULL;
 	
 	if (ZEND_NUM_ARGS()!=1 || zend_get_parameters_ex(1, &arg)==FAILURE) {
 		ZEND_WRONG_PARAM_COUNT();
 	}
-	if (((*arg)->type != IS_OBJECT) || !(*arg)->value.obj.ce->parent) {
+
+	if (Z_TYPE_PP(arg) == IS_OBJECT)
+	   ce = Z_OBJCE_PP(arg);
+	else if (Z_TYPE_PP(arg) == IS_STRING) {
+		SEPARATE_ZVAL(arg);
+		zend_str_tolower(Z_STRVAL_PP(arg), Z_STRLEN_PP(arg));
+		zend_hash_find(EG(class_table), Z_STRVAL_PP(arg), Z_STRLEN_PP(arg)+1, (void **)&ce);
+	}
+
+	if (ce && ce->parent) {
+		RETURN_STRINGL(ce->parent->name, ce->parent->name_length, 1);
+	} else {
 		RETURN_FALSE;
 	}
-	RETURN_STRINGL((*arg)->value.obj.ce->parent->name,	(*arg)->value.obj.ce->parent->name_length, 1);
 }
 /* }}} */
 
-/* {{{ proto bool is_subclass_of(object object, string class_name)
-   Returns true if the object has this class as one of its parents */
-ZEND_FUNCTION(is_subclass_of)
+static void is_a_impl(INTERNAL_FUNCTION_PARAMETERS, zend_bool only_subclass)
 {
 	zval **obj, **class_name;
 	char *lcname;
-	zend_class_entry *parent_ce = NULL;
+	zend_class_entry *ce = NULL;
 
 	if (ZEND_NUM_ARGS() != 2 || zend_get_parameters_ex(2, &obj, &class_name)==FAILURE) {
 		ZEND_WRONG_PARAM_COUNT();
 	}
 
-	if ((*obj)->type != IS_OBJECT) {
+	if (Z_TYPE_PP(obj) != IS_OBJECT) {
 		RETURN_FALSE;
 	}
 
 	convert_to_string_ex(class_name);
-	lcname = estrndup((*class_name)->value.str.val, (*class_name)->value.str.len);
-	zend_str_tolower(lcname, (*class_name)->value.str.len);
+	lcname = estrndup(Z_STRVAL_PP(class_name), Z_STRLEN_PP(class_name));
+	zend_str_tolower(lcname, Z_STRLEN_PP(class_name));
 
-	for (parent_ce = (*obj)->value.obj.ce->parent; parent_ce != NULL; parent_ce = parent_ce->parent) {
-		if (!strcmp(parent_ce->name, lcname)) {
+	if (only_subclass)
+		ce = Z_OBJCE_PP(obj)->parent;
+	else
+		ce = Z_OBJCE_PP(obj);
+	for (; ce != NULL; ce = ce->parent) {
+		if ((ce->name_length == (uint)Z_STRLEN_PP(class_name)) && !zend_binary_strcasecmp(ce->name, ce->name_length, lcname, Z_STRLEN_PP(class_name))) {
 			efree(lcname);
 			RETURN_TRUE;
 		}
@@ -505,8 +576,25 @@ ZEND_FUNCTION(is_subclass_of)
 	efree(lcname);
 	RETURN_FALSE;
 }
+
+/* {{{ proto bool is_subclass_of(object object, string class_name)
+   Returns true if the object has this class as one of its parents */
+ZEND_FUNCTION(is_subclass_of)
+{
+	is_a_impl(INTERNAL_FUNCTION_PARAM_PASSTHRU, 1);
+}
 /* }}} */
 
+
+/* {{{ proto bool is_a(object object, string class_name)
+   Returns true if the object is of this class or has this class as one of its parents */
+ZEND_FUNCTION(is_a)
+{
+	is_a_impl(INTERNAL_FUNCTION_PARAM_PASSTHRU, 0);
+}
+/* }}} */
+
+
 /* {{{ proto array get_class_vars(string class_name)
    Returns an array of default properties of the class */
 ZEND_FUNCTION(get_class_vars)
@@ -515,7 +603,6 @@ ZEND_FUNCTION(get_class_vars)
 	char *lcname;
 	zend_class_entry *ce;
 	zval *tmp;
-	CLS_FETCH();
 
 	if (ZEND_NUM_ARGS() != 1 || zend_get_parameters_ex(1, &class_name)==FAILURE) {
 		ZEND_WRONG_PARAM_COUNT();
@@ -525,14 +612,14 @@ ZEND_FUNCTION(get_class_vars)
 	lcname = estrndup((*class_name)->value.str.val, (*class_name)->value.str.len);
 	zend_str_tolower(lcname, (*class_name)->value.str.len);
 
-	if (zend_hash_find(CG(class_table), lcname, (*class_name)->value.str.len+1, (void **)&ce)==FAILURE) {
+	if (zend_hash_find(EG(class_table), lcname, (*class_name)->value.str.len+1, (void **)&ce)==FAILURE) {
 		efree(lcname);
 		RETURN_FALSE;
 	} else {
 		efree(lcname);
 		array_init(return_value);
 		if (!ce->constants_updated) {
-			zend_hash_apply_with_argument(&ce->default_properties, (int (*)(void *,void *)) zval_update_constant, (void *) 1);
+			zend_hash_apply_with_argument(&ce->default_properties, (apply_func_arg_t) zval_update_constant, (void *) 1 TSRMLS_CC);
 			ce->constants_updated = 1;
 		}
 		zend_hash_copy(return_value->value.ht, &ce->default_properties, (copy_ctor_func_t) zval_add_ref, (void *) &tmp, sizeof(zval *));
@@ -540,6 +627,7 @@ ZEND_FUNCTION(get_class_vars)
 }
 /* }}} */
 
+
 /* {{{ proto array get_object_vars(object obj)
    Returns an array of object properties */
 ZEND_FUNCTION(get_object_vars)
@@ -556,51 +644,53 @@ ZEND_FUNCTION(get_object_vars)
 	}
 
 	array_init(return_value);
-	zend_hash_copy(return_value->value.ht, (*obj)->value.obj.properties,
+	zend_hash_copy(return_value->value.ht, Z_OBJPROP_PP(obj),
 				   (copy_ctor_func_t) zval_add_ref, (void *) &tmp, sizeof(zval *));
 }
 /* }}} */
 
-/* {{{ proto array get_class_methods(string class_name)
-   Returns an array of class methods' names */
+
+/* {{{ proto array get_class_methods(mixed class)
+   Returns an array of method names for class or class instance. */
 ZEND_FUNCTION(get_class_methods)
 {
-	zval **class_name;
+	zval **class;
 	zval *method_name;
-	char *lcname;
-	zend_class_entry *ce;
+	zend_class_entry *ce = NULL;
 	char *string_key;
 	ulong num_key;
 	int key_type;
-	CLS_FETCH();
 
-	if (ZEND_NUM_ARGS() != 1 || zend_get_parameters_ex(1, &class_name)==FAILURE) {
+	if (ZEND_NUM_ARGS() != 1 || zend_get_parameters_ex(1, &class)==FAILURE) {
 		ZEND_WRONG_PARAM_COUNT();
 	}
 
-	convert_to_string_ex(class_name);
-	lcname = estrndup((*class_name)->value.str.val, (*class_name)->value.str.len);
-	zend_str_tolower(lcname, (*class_name)->value.str.len);
+	if (Z_TYPE_PP(class) == IS_OBJECT)
+		ce = Z_OBJCE_PP(class);
+	else if (Z_TYPE_PP(class) == IS_STRING) {
+		SEPARATE_ZVAL(class);
+		zend_str_tolower(Z_STRVAL_PP(class), Z_STRLEN_PP(class));
+		zend_hash_find(EG(class_table), Z_STRVAL_PP(class), Z_STRLEN_PP(class)+1, (void **)&ce);
+	}
 
-	if (zend_hash_find(CG(class_table), lcname, (*class_name)->value.str.len+1, (void **)&ce)==FAILURE) {
-		efree(lcname);
+	if (!ce) {
 		RETURN_NULL();
-	} else {
-		efree(lcname);
-		array_init(return_value);
-		zend_hash_internal_pointer_reset(&ce->function_table);
-		while ((key_type = zend_hash_get_current_key(&ce->function_table, &string_key, &num_key, 1)) != HASH_KEY_NON_EXISTANT) {
-			if (key_type == HASH_KEY_IS_STRING) {
-				MAKE_STD_ZVAL(method_name);
-				ZVAL_STRING(method_name, string_key, 0);
-				zend_hash_next_index_insert(return_value->value.ht, &method_name, sizeof(zval *), NULL);
-			}
-			zend_hash_move_forward(&ce->function_table);
+	}
+
+	array_init(return_value);
+	zend_hash_internal_pointer_reset(&ce->function_table);
+	while ((key_type = zend_hash_get_current_key(&ce->function_table, &string_key, &num_key, 1)) != HASH_KEY_NON_EXISTANT) {
+		if (key_type == HASH_KEY_IS_STRING) {
+			MAKE_STD_ZVAL(method_name);
+			ZVAL_STRING(method_name, string_key, 0);
+			zend_hash_next_index_insert(return_value->value.ht, &method_name, sizeof(zval *), NULL);
 		}
+		zend_hash_move_forward(&ce->function_table);
 	}
 }
 /* }}} */
 
+
 /* {{{ proto bool method_exists(object object, string method)
    Checks if the class method exists */
 ZEND_FUNCTION(method_exists)
@@ -617,7 +707,7 @@ ZEND_FUNCTION(method_exists)
 	convert_to_string_ex(method_name);
 	lcname = estrndup((*method_name)->value.str.val, (*method_name)->value.str.len);
 	zend_str_tolower(lcname, (*method_name)->value.str.len);
-	if(zend_hash_exists(&(*klass)->value.obj.ce->function_table, lcname, (*method_name)->value.str.len+1)) {
+	if (zend_hash_exists(&Z_OBJCE_PP(klass)->function_table, lcname, (*method_name)->value.str.len+1)) {
 		efree(lcname);
 		RETURN_TRUE;
 	} else {
@@ -627,13 +717,13 @@ ZEND_FUNCTION(method_exists)
 }
 /* }}} */
 
+
 /* {{{ proto bool class_exists(string classname)
    Checks if the class exists */
 ZEND_FUNCTION(class_exists)
 {
 	zval **class_name;
 	char *lcname;
-	CLS_FETCH();
 
 	if (ZEND_NUM_ARGS() != 1 || zend_get_parameters_ex(1, &class_name)==FAILURE) {
 		ZEND_WRONG_PARAM_COUNT();
@@ -641,7 +731,7 @@ ZEND_FUNCTION(class_exists)
 	convert_to_string_ex(class_name);
 	lcname = estrndup((*class_name)->value.str.val, (*class_name)->value.str.len);
 	zend_str_tolower(lcname, (*class_name)->value.str.len);
-	if (zend_hash_exists(CG(class_table), lcname, (*class_name)->value.str.len+1)) {
+	if (zend_hash_exists(EG(class_table), lcname, (*class_name)->value.str.len+1)) {
 		efree(lcname);
 		RETURN_TRUE;
 	} else {
@@ -651,13 +741,15 @@ ZEND_FUNCTION(class_exists)
 }
 /* }}} */
 
+
 /* {{{ proto bool function_exists(string function_name) 
    Checks if the function exists */
 ZEND_FUNCTION(function_exists)
 {
 	zval **function_name;
+	zend_function *func;
 	char *lcname;
-	int retval;
+	zend_bool retval;
 	
 	if (ZEND_NUM_ARGS()!=1 || zend_get_parameters_ex(1, &function_name)==FAILURE) {
 		ZEND_WRONG_PARAM_COUNT();
@@ -666,14 +758,25 @@ ZEND_FUNCTION(function_exists)
 	lcname = estrndup((*function_name)->value.str.val, (*function_name)->value.str.len);
 	zend_str_tolower(lcname, (*function_name)->value.str.len);
 
-	retval = zend_hash_exists(EG(function_table), lcname, (*function_name)->value.str.len+1);
+	retval = (zend_hash_find(EG(function_table), lcname, (*function_name)->value.str.len+1, (void **)&func) == SUCCESS);
 	efree(lcname);
 
+	/*
+	 * A bit of a hack, but not a bad one: we see if the handler of the function
+	 * is actually one that displays "function is disabled" message.
+	 */
+	if (retval &&
+		func->internal_function.handler == zif_display_disabled_function) {
+		retval = 0;
+	}
+
 	RETURN_BOOL(retval);
 }
 /* }}} */
 
-
+#if ZEND_DEBUG
+/* {{{ proto void leak(int num_bytes=3)
+   Cause an intentional memory leak, for testing/debugging purposes */
 ZEND_FUNCTION(leak)
 {
 	int leakbytes=3;
@@ -688,6 +791,7 @@ ZEND_FUNCTION(leak)
 	
 	emalloc(leakbytes);
 }
+/* }}} */
 
 
 #ifdef ZEND_TEST_EXCEPTIONS
@@ -699,26 +803,7 @@ ZEND_FUNCTION(crash)
 }
 #endif
 
-
-static int copy_import_use_file(zend_file_handle *fh, zval *array)
-{
-	if (fh->filename) {
-		char *extension_start;
-
-		extension_start = strstr(fh->filename, zend_uv.import_use_extension);
-		if (extension_start) {
-			*extension_start = 0;
-			if (fh->opened_path) {
-				add_assoc_string(array, fh->filename, fh->opened_path, 1);
-			} else {
-				add_assoc_stringl(array, fh->filename, "N/A", sizeof("N/A")-1, 1);
-			}
-			*extension_start = zend_uv.import_use_extension[0];
-		}
-	}
-	return 0;
-}
-
+#endif /* ZEND_DEBUG */
 
 /* {{{ proto array get_included_files(void)
    Returns an array with the file names that were include_once()'d */
@@ -731,12 +816,10 @@ ZEND_FUNCTION(get_included_files)
 
 	array_init(return_value);
 	zend_hash_internal_pointer_reset(&EG(included_files));
-	while(zend_hash_get_current_key(&EG(included_files), &entry, NULL, 1) == HASH_KEY_IS_STRING) {
-		add_next_index_string(return_value,entry,0);
+	while (zend_hash_get_current_key(&EG(included_files), &entry, NULL, 1) == HASH_KEY_IS_STRING) {
+		add_next_index_string(return_value, entry, 0);
 		zend_hash_move_forward(&EG(included_files));
 	}
-
-	/*	zend_hash_apply_with_argument(&EG(included_files), (apply_func_arg_t) copy_import_use_file, return_value); */
 }
 /* }}} */
 
@@ -786,13 +869,21 @@ ZEND_FUNCTION(trigger_error)
 ZEND_FUNCTION(set_error_handler)
 {
 	zval **error_handler;
+	char *error_handler_name;
 	zend_bool had_orig_error_handler=0;
 
 	if (ZEND_NUM_ARGS()!=1 || zend_get_parameters_ex(1, &error_handler)==FAILURE) {
 		ZEND_WRONG_PARAM_COUNT();
 	}
 
-	convert_to_string_ex(error_handler);
+	if (!zend_is_callable(*error_handler, 0, &error_handler_name)) {
+		zend_error(E_WARNING, "%s() expects argument 1, '%s', to be a valid callback",
+				   get_active_function_name(TSRMLS_C), error_handler_name);
+		efree(error_handler_name);
+		return;
+	}
+	efree(error_handler_name);
+
 	if (EG(user_error_handler)) {
 		had_orig_error_handler = 1;
 		*return_value = *EG(user_error_handler);
@@ -801,7 +892,7 @@ ZEND_FUNCTION(set_error_handler)
 	}
 	ALLOC_ZVAL(EG(user_error_handler));
 
-	if (Z_STRLEN_PP(error_handler)==0) { /* unset user-defined handler */
+	if (!zend_is_true(*error_handler)) { /* unset user-defined handler */
 		FREE_ZVAL(EG(user_error_handler));
 		EG(user_error_handler) = NULL;
 		RETURN_TRUE;
@@ -817,7 +908,6 @@ ZEND_FUNCTION(set_error_handler)
 /* }}} */
 
 
-
 /* {{{ proto void restore_error_handler(void)
    Restores the previously defined error handler function */
 ZEND_FUNCTION(restore_error_handler)
@@ -844,21 +934,21 @@ static int copy_class_name(zend_class_entry *ce, int num_args, va_list args, zen
 	return 0;
 }
 
+
 /* {{{ proto array get_declared_classes(void)
    Returns an array of all declared classes. */
 ZEND_FUNCTION(get_declared_classes)
 {
-	CLS_FETCH();
-
 	if (ZEND_NUM_ARGS() != 0) {
 		ZEND_WRONG_PARAM_COUNT();
 	}
 
 	array_init(return_value);
-	zend_hash_apply_with_arguments(CG(class_table), (apply_func_args_t)copy_class_name, 1, return_value);
+	zend_hash_apply_with_arguments(EG(class_table), (apply_func_args_t) copy_class_name, 1, return_value);
 }
 /* }}} */
 
+
 static int copy_function_name(zend_function *func, int num_args, va_list args, zend_hash_key *hash_key)
 {
 	zval *internal_ar = va_arg(args, zval *),
@@ -877,6 +967,7 @@ static int copy_function_name(zend_function *func, int num_args, va_list args, z
 	return 0;
 }
 
+
 /* {{{ proto array get_defined_functions(void)
    Returns an array of all defined functions */
 ZEND_FUNCTION(get_defined_functions)
@@ -895,7 +986,7 @@ ZEND_FUNCTION(get_defined_functions)
 	array_init(user);
 	array_init(return_value);
 	
-	zend_hash_apply_with_arguments(EG(function_table), (apply_func_args_t)copy_function_name, 2, internal, user);
+	zend_hash_apply_with_arguments(EG(function_table), (apply_func_args_t) copy_function_name, 2, internal, user);
 	
 	if (zend_hash_add(return_value->value.ht, "internal", sizeof("internal"), (void **)&internal, sizeof(zval *), NULL) == FAILURE) {
 		zend_error(E_WARNING, "Cannot add internal functions to return value from get_defined_functions()");
@@ -909,6 +1000,7 @@ ZEND_FUNCTION(get_defined_functions)
 }
 /* }}} */
 
+
 /* {{{ proto array get_defined_vars(void)
    Returns an associative array of names and values of all currently defined variable names (variables in the current scope) */
 ZEND_FUNCTION(get_defined_vars)
@@ -922,8 +1014,8 @@ ZEND_FUNCTION(get_defined_vars)
 }
 /* }}} */
 
-#define LAMBDA_TEMP_FUNCNAME	"__lambda_func"
 
+#define LAMBDA_TEMP_FUNCNAME	"__lambda_func"
 /* {{{ proto string create_function(string args, string code)
    Creates an anonymous function, and returns its name (funny, eh?) */
 ZEND_FUNCTION(create_function)
@@ -933,7 +1025,6 @@ ZEND_FUNCTION(create_function)
 	zval **z_function_args, **z_function_code;
 	int retval;
 	char *eval_name;
-	CLS_FETCH();
 
 	if (ZEND_NUM_ARGS()!=2 || zend_get_parameters_ex(2, &z_function_args, &z_function_code)==FAILURE) {
 		ZEND_WRONG_PARAM_COUNT();
@@ -951,26 +1042,27 @@ ZEND_FUNCTION(create_function)
 	eval_code = (char *) emalloc(eval_code_length);
 	sprintf(eval_code, "function " LAMBDA_TEMP_FUNCNAME "(%s){%s}", Z_STRVAL_PP(z_function_args), Z_STRVAL_PP(z_function_code));
 
-	eval_name = zend_make_compiled_string_description("runtime-created function");
-	retval = zend_eval_string(eval_code, NULL, eval_name CLS_CC ELS_CC);
+	eval_name = zend_make_compiled_string_description("runtime-created function" TSRMLS_CC);
+	retval = zend_eval_string(eval_code, NULL, eval_name TSRMLS_CC);
 	efree(eval_code);
 	efree(eval_name);
 
 	if (retval==SUCCESS) {
-		zend_function *func;
+		zend_function *func, new_function;
 
 		if (zend_hash_find(EG(function_table), LAMBDA_TEMP_FUNCNAME, sizeof(LAMBDA_TEMP_FUNCNAME), (void **) &func)==FAILURE) {
 			zend_error(E_ERROR, "Unexpected inconsistency in create_function()");
 			RETURN_FALSE;
 		}
-		function_add_ref(func);
+		new_function = *func;
+		function_add_ref(&new_function);
 
 		function_name = (char *) emalloc(sizeof("0lambda_")+MAX_LENGTH_OF_LONG);
 
 		do {
 			sprintf(function_name, "%clambda_%d", 0, ++EG(lambda_count));
 			function_name_length = strlen(function_name+1)+1;
-		} while (zend_hash_add(EG(function_table), function_name, function_name_length+1, func, sizeof(zend_function), NULL)==FAILURE);
+		} while (zend_hash_add(EG(function_table), function_name, function_name_length+1, &new_function, sizeof(zend_function), NULL)==FAILURE);
 		zend_hash_del(EG(function_table), LAMBDA_TEMP_FUNCNAME, sizeof(LAMBDA_TEMP_FUNCNAME));
 		RETURN_STRINGL(function_name, function_name_length, 0);
 	} else {
@@ -989,7 +1081,8 @@ ZEND_FUNCTION(zend_test_func)
 }
 #endif
 
-
+/* {{{ proto string get_resource_type(resource res)
+   Get the resource type name for a given resource */
 ZEND_FUNCTION(get_resource_type)
 {
 	char *resource_type;
@@ -1004,10 +1097,273 @@ ZEND_FUNCTION(get_resource_type)
 		RETURN_FALSE;
 	}
 
-	resource_type = zend_rsrc_list_get_rsrc_type(Z_LVAL_PP(z_resource_type));
+	resource_type = zend_rsrc_list_get_rsrc_type(Z_LVAL_PP(z_resource_type) TSRMLS_CC);
 	if (resource_type) {
 		RETURN_STRING(resource_type, 1);
 	} else {
 		RETURN_STRING("Unknown", 1);
 	}
 }
+/* }}} */
+
+static int add_extension_info(zend_module_entry *module, void *arg TSRMLS_DC)
+{
+	zval *name_array = (zval *)arg;
+	add_next_index_string(name_array, module->name, 1);
+	return 0;
+}
+
+static int add_constant_info(zend_constant *constant, void *arg TSRMLS_DC)
+{
+	zval *name_array = (zval *)arg;
+	zval *const_val;
+
+	MAKE_STD_ZVAL(const_val);
+	*const_val = constant->value;
+	zval_copy_ctor(const_val);
+	INIT_PZVAL(const_val);
+	add_assoc_zval_ex(name_array, constant->name, constant->name_len, const_val);
+	return 0;
+}
+
+
+/* {{{ proto array get_loaded_extensions(void)
+   Return an array containing names of loaded extensions */
+ZEND_FUNCTION(get_loaded_extensions)
+{
+	if (ZEND_NUM_ARGS() != 0) {
+		ZEND_WRONG_PARAM_COUNT();
+	}
+
+	array_init(return_value);
+	zend_hash_apply_with_argument(&module_registry, (apply_func_arg_t) add_extension_info, return_value TSRMLS_CC);
+}
+/* }}} */
+
+
+/* {{{ proto array get_defined_constants(void)
+   Return an array containing the names and values of all defined constants */
+ZEND_FUNCTION(get_defined_constants)
+{
+	if (ZEND_NUM_ARGS() != 0) {
+		ZEND_WRONG_PARAM_COUNT();
+	}
+
+	array_init(return_value);
+	zend_hash_apply_with_argument(EG(zend_constants), (apply_func_arg_t) add_constant_info, return_value TSRMLS_CC);
+}
+
+/* }}} */
+
+static zval *debug_backtrace_get_args(void ***curpos TSRMLS_DC) {
+	void **p = *curpos - 2;
+	zval *arg_array, **arg;
+	int arg_count = (ulong) *p;
+
+	*curpos -= (arg_count+2); 
+
+	MAKE_STD_ZVAL(arg_array);
+	array_init(arg_array);
+	p -= arg_count;
+
+	while (--arg_count >= 0) {
+		arg = (zval **) p++;
+		SEPARATE_ZVAL_TO_MAKE_IS_REF(arg);
+		(*arg)->refcount++;
+		add_next_index_zval(arg_array, *arg);
+	}
+	return arg_array;
+}
+
+/* {{{ proto void debug_backtrace(void)
+   Prints out a backtrace */
+ZEND_FUNCTION(debug_backtrace)
+{
+	zend_execute_data *ptr;
+	int lineno;
+	char *function_name;
+	char *filename;
+	char *class_name;
+	char *call_type;
+	char *include_filename = NULL;
+	zval *stack_frame;
+	void **cur_arg_pos = EG(argument_stack).top_element;
+	void **args = cur_arg_pos;
+	int arg_stack_consistent = 0;
+	int frames_on_stack = 0;
+
+	if (ZEND_NUM_ARGS()) {
+		ZEND_WRONG_PARAM_COUNT();
+	}
+
+	while (--args >= EG(argument_stack).elements) {
+		if (*args--) {
+			break;
+		}
+		args -= *(ulong*)args;
+		frames_on_stack++;
+
+		if (args == EG(argument_stack).elements) {
+			arg_stack_consistent = 1;
+			break;
+		}
+	}
+
+	ptr = EG(current_execute_data);
+
+	/* skip debug_backtrace() */
+	ptr = ptr->prev_execute_data;
+	cur_arg_pos -= 2;
+	frames_on_stack--;
+
+	array_init(return_value);
+
+	while (ptr) {
+		MAKE_STD_ZVAL(stack_frame);
+		array_init(stack_frame);
+
+		if (ptr->op_array) {
+			filename = ptr->op_array->filename;
+			lineno = ptr->opline->lineno;
+			add_assoc_string_ex(stack_frame, "file", sizeof("file"), filename, 1);
+			add_assoc_long_ex(stack_frame, "line", sizeof("line"), lineno);
+
+			/* try to fetch args only if an FCALL was just made - elsewise we're in the middle of a function
+			 * and debug_baktrace() might have been called by the error_handler. in this case we don't 
+			 * want to pop anything of the argument-stack */
+		} else {
+			filename = NULL;
+		}
+
+		function_name = ptr->function_state.function->type != ZEND_OVERLOADED_FUNCTION ? ptr->function_state.function->common.function_name : get_active_function_name(TSRMLS_C);  /* "OVERLOADED_FUNCTION" */
+
+		if (function_name) {
+			add_assoc_string_ex(stack_frame, "function", sizeof("function"), function_name, 1);
+
+			if (ptr->ce) {
+				class_name = ptr->ce->name;
+				call_type = "::";
+			} else if (ptr->object.ptr && ptr->object.ptr->type == IS_OBJECT && ptr->object.ptr->value.obj.ce) {
+				class_name = ptr->object.ptr->value.obj.ce->name;
+				call_type = "->";
+			} else {
+				class_name = NULL;
+				call_type = NULL;
+			}
+
+			if (class_name) {
+				add_assoc_string_ex(stack_frame, "class", sizeof("class"), class_name, 1);
+				add_assoc_string_ex(stack_frame, "type", sizeof("type"), call_type, 1);
+			}
+
+			if ((! ptr->opline) || ((ptr->opline->opcode == ZEND_DO_FCALL_BY_NAME) || (ptr->opline->opcode == ZEND_DO_FCALL))) {
+				if (arg_stack_consistent && (frames_on_stack > 0)) {
+					add_assoc_zval_ex(stack_frame, "args", sizeof("args"), debug_backtrace_get_args(&cur_arg_pos TSRMLS_CC));
+					frames_on_stack--;
+				}
+			}	
+		} else {
+			/* i know this is kinda ugly, but i'm trying to avoid extra cycles in the main execution loop */
+			zend_bool build_filename_arg = 1;
+
+			switch (ptr->opline->op2.u.constant.value.lval) {
+				case ZEND_EVAL:
+					function_name = "eval";
+					build_filename_arg = 0;
+					break;
+				case ZEND_INCLUDE:
+					function_name = "include";
+					break;
+				case ZEND_REQUIRE:
+					function_name = "require";
+					break;
+				case ZEND_INCLUDE_ONCE:
+					function_name = "include_once";
+					break;
+				case ZEND_REQUIRE_ONCE:
+					function_name = "require_once";
+					break;
+				default:
+					/* this can actually happen if you use debug_backtrace() in your error_handler and 
+					 * you're in the top-scope */
+					function_name = "unknown"; 
+					build_filename_arg = 0;
+					break;
+			}
+
+			if (build_filename_arg && include_filename) {
+				zval *arg_array;
+
+				MAKE_STD_ZVAL(arg_array);
+				array_init(arg_array);
+				
+				/* include_filename always points to the last filename of the last last called-fuction.
+				   if we have called include in the frame above - this is the file we have included.
+				 */
+
+				add_next_index_string(arg_array, include_filename, 1);
+				add_assoc_zval_ex(stack_frame, "args", sizeof("args"), arg_array);
+			}
+
+			add_assoc_string_ex(stack_frame, "function", sizeof("function"), function_name, 1);
+		}
+
+		add_next_index_zval(return_value, stack_frame);
+
+		include_filename = filename; 
+
+		ptr = ptr->prev_execute_data;
+	}
+}
+
+
+/* {{{ proto bool extension_loaded(string extension_name)
+   Returns true if the named extension is loaded */
+ZEND_FUNCTION(extension_loaded)
+{
+	zval **extension_name;
+
+	if (ZEND_NUM_ARGS() != 1 || zend_get_parameters_ex(1, &extension_name)) {
+		ZEND_WRONG_PARAM_COUNT();
+	}
+
+	convert_to_string_ex(extension_name);
+	if (zend_hash_exists(&module_registry, Z_STRVAL_PP(extension_name), Z_STRLEN_PP(extension_name)+1)) {
+		RETURN_TRUE;
+	} else {
+		RETURN_FALSE;
+	}
+}
+/* }}} */
+
+
+/* {{{ proto array get_extension_funcs(string extension_name)
+   Returns an array with the names of functions belonging to the named extension */
+ZEND_FUNCTION(get_extension_funcs)
+{
+	zval **extension_name;
+	zend_module_entry *module;
+	zend_function_entry *func;
+
+	if (ZEND_NUM_ARGS() != 1 || zend_get_parameters_ex(1, &extension_name)) {
+		ZEND_WRONG_PARAM_COUNT();
+	}
+
+	convert_to_string_ex(extension_name);
+	if (zend_hash_find(&module_registry, Z_STRVAL_PP(extension_name),
+		Z_STRLEN_PP(extension_name)+1, (void**)&module) == FAILURE) {
+		return;
+	}
+
+	array_init(return_value);
+	func = module->functions;
+	if (!func) {
+		return;
+	}
+
+	while (func->fname) {
+		add_next_index_string(return_value, func->fname, 1);
+		func++;
+	}
+}
+/* }}} */
