@@ -2,12 +2,12 @@
    +----------------------------------------------------------------------+
    | Zend Engine                                                          |
    +----------------------------------------------------------------------+
-   | Copyright (c) 1998-2001 Zend Technologies Ltd. (http://www.zend.com) |
+   | Copyright (c) 1998-2003 Zend Technologies Ltd. (http://www.zend.com) |
    +----------------------------------------------------------------------+
-   | This source file is subject to version 0.92 of the Zend license,     |
+   | This source file is subject to version 2.00 of the Zend license,     |
    | that is bundled with this package in the file LICENSE, and is        | 
    | available at through the world-wide-web at                           |
-   | http://www.zend.com/license/0_92.txt.                                |
+   | http://www.zend.com/license/2_00.txt.                                |
    | If you did not receive a copy of the Zend license and are unable to  |
    | obtain it through the world-wide-web, please send a note to          |
    | license@zend.com so we can mail you a copy immediately.              |
@@ -53,22 +53,43 @@ ZEND_API void zend_html_putc(char c)
 }
 
 
-ZEND_API void zend_html_puts(char *s, uint len)
+ZEND_API void zend_html_puts(const char *s, uint len TSRMLS_DC)
 {
-	register char *ptr=s, *end=s+len;
+	const char *ptr=s, *end=s+len;
+#ifdef ZEND_MULTIBYTE
+	char *mbs;
+	int mblen;
+
+	if (LANG_SCNG(output_filter)) {
+		LANG_SCNG(output_filter)(&mbs, &mblen, s, len TSRMLS_CC);
+		ptr = mbs;
+		end = mbs+mblen;
+	}
+#endif /* ZEND_MULTIBYTE */
 	
 	while (ptr<end) {
-		if (*ptr==' '
-			&& len>1
-			&& !(((ptr+1)>=end) || (*(ptr+1)==' ')) /* next is not a space */
-			&& !((ptr==s) || (*(ptr-1)==' '))) /* last is not a space */ {
-			char c = *ptr++;
-
-			ZEND_PUTC(c);
-			continue;
+		if (*ptr==' ') {
+			/* Series of spaces should be displayed as &nbsp;'s
+			 * whereas single spaces should be displayed as a space
+			 */
+			if ((ptr+1) < end && *(ptr+1)==' ') {
+				do {
+					zend_html_putc(*ptr);
+				} while ((++ptr < end) && (*ptr==' '));
+			} else {
+				(void) ZEND_PUTC(*ptr);
+				ptr++;
+			}
+		} else {
+			zend_html_putc(*ptr++);
 		}
-		zend_html_putc(*ptr++);
 	}
+
+#ifdef ZEND_MULTIBYTE
+	if (LANG_SCNG(output_filter)) {
+		efree(mbs);
+	}
+#endif /* ZEND_MULTIBYTE */
 }
 
 
@@ -108,7 +129,7 @@ ZEND_API void zend_highlight(zend_syntax_highlighter_ini *syntax_highlighter_ini
 				in_string = !in_string;
 				break;				
 			case T_WHITESPACE:
-				zend_html_puts(LANG_SCNG(yy_text), LANG_SCNG(yy_leng));  /* no color needed */
+				zend_html_puts(LANG_SCNG(yy_text), LANG_SCNG(yy_leng) TSRMLS_CC);  /* no color needed */
 				token.type = 0;
 				continue;
 				break;
@@ -136,10 +157,10 @@ ZEND_API void zend_highlight(zend_syntax_highlighter_ini *syntax_highlighter_ini
 		}
 		switch (token_type) {
 			case T_END_HEREDOC:
-				zend_html_puts(token.value.str.val, token.value.str.len);
+				zend_html_puts(token.value.str.val, token.value.str.len TSRMLS_CC);
 				break;
 			default:
-				zend_html_puts(LANG_SCNG(yy_text), LANG_SCNG(yy_leng));
+				zend_html_puts(LANG_SCNG(yy_text), LANG_SCNG(yy_leng) TSRMLS_CC);
 				break;
 		}
 
@@ -173,9 +194,74 @@ ZEND_API void zend_highlight(zend_syntax_highlighter_ini *syntax_highlighter_ini
 	zend_printf("</code>");
 }
 
+
+ZEND_API void zend_strip(TSRMLS_D)
+{
+	zval token;
+	int token_type;
+	int prev_space = 0;
+
+	token.type = 0;
+	while ((token_type=lex_scan(&token TSRMLS_CC))) {
+		switch (token_type) {
+			case T_WHITESPACE:
+				if (!prev_space) {
+					putchar(' ');
+					prev_space = 1;
+				}
+						/* lack of break; is intentional */
+			case T_COMMENT:
+				token.type = 0;
+				continue;
+			
+			case T_END_HEREDOC: {
+					char *ptr = LANG_SCNG(yy_text);
+
+					fwrite(ptr, LANG_SCNG(yy_leng) - 1, 1, stdout);
+					/* The ensure that we only write one ; and that it followed by the required newline */
+					putchar('\n');
+					if (ptr[LANG_SCNG(yy_leng) - 1] == ';') {
+						lex_scan(&token TSRMLS_CC);
+					}
+					efree(token.value.str.val);
+				}
+				break;
+			
+			default:
+				fwrite(LANG_SCNG(yy_text), LANG_SCNG(yy_leng), 1, stdout);
+				break;
+		}
+
+		if (token.type == IS_STRING) {
+			switch (token_type) {
+				case T_OPEN_TAG:
+				case T_OPEN_TAG_WITH_ECHO:
+				case T_CLOSE_TAG:
+				case T_WHITESPACE:
+				case T_COMMENT:
+					break;
+
+				default:
+					efree(token.value.str.val);
+					break;
+			}
+		}
+		prev_space = token.type = 0;
+	}
+#ifdef ZEND_MULTIBYTE
+	if (LANG_SCNG(code)) {
+		efree(LANG_SCNG(code));
+	}
+	if (LANG_SCNG(current_code)) {
+		efree(LANG_SCNG(current_code));
+	}
+#endif /* ZEND_MULTIBYTE */
+}
+
 /*
  * Local variables:
  * tab-width: 4
  * c-basic-offset: 4
  * End:
  */
+
