@@ -1,8 +1,8 @@
 /*
    +----------------------------------------------------------------------+
-   | PHP version 4.0                                                      |
+   | PHP Version 4                                                        |
    +----------------------------------------------------------------------+
-   | Copyright (c) 1997-2001 The PHP Group                                |
+   | Copyright (c) 1997-2003 The PHP Group                                |
    +----------------------------------------------------------------------+
    | This source file is subject to version 2.02 of the PHP license,      |
    | that is bundled with this package in the file LICENSE, and is        |
@@ -12,11 +12,11 @@
    | obtain it through the world-wide-web, please send a note to          |
    | license@php.net so we can mail you a copy immediately.               |
    +----------------------------------------------------------------------+
-   | Authors: Andrei Zmievski <andrei@ispi.net>                           |
+   | Author: Andrei Zmievski <andrei@php.net>                             |
    +----------------------------------------------------------------------+
  */
 
-/* $Id: php_pcre.c,v 1.108.2.2 2001/11/10 23:50:53 andrei Exp $ */
+/* $Id: php_pcre.c,v 1.132.2.10 2003/09/12 01:32:38 sniper Exp $ */
 
 #ifdef HAVE_CONFIG_H
 #include "config.h"
@@ -32,14 +32,18 @@
 
 #include "ext/standard/php_string.h"
 
-#define PREG_PATTERN_ORDER			0
-#define PREG_SET_ORDER				1
+#define PREG_PATTERN_ORDER			1
+#define PREG_SET_ORDER				2
+#define PREG_OFFSET_CAPTURE			(1<<8)
 
 #define	PREG_SPLIT_NO_EMPTY			(1<<0)
 #define PREG_SPLIT_DELIM_CAPTURE	(1<<1)
+#define PREG_SPLIT_OFFSET_CAPTURE	(1<<2)
 
 #define PREG_REPLACE_EVAL			(1<<0)
 
+#define PREG_GREP_INVERT			(1<<0)
+
 
 ZEND_DECLARE_MODULE_GLOBALS(pcre)
 
@@ -52,6 +56,7 @@ static void *php_pcre_malloc(size_t size)
 
 static void php_pcre_free(void *ptr)
 {
+	if (ptr)
 	pefree(ptr, 1);
 }
 
@@ -59,9 +64,11 @@ static void php_pcre_free(void *ptr)
 static void php_free_pcre_cache(void *data)
 {
 	pcre_cache_entry *pce = (pcre_cache_entry *) data;
+	if (!pce) return;
 	pefree(pce->re, 1);
 #if HAVE_SETLOCALE
 	if ((void*)pce->tables) pefree((void*)pce->tables, 1);
+	pefree(pce->locale, 1);
 #endif
 }
 
@@ -71,12 +78,10 @@ static void php_pcre_init_globals(zend_pcre_globals *pcre_globals TSRMLS_DC)
 	zend_hash_init(&pcre_globals->pcre_cache, 0, NULL, php_free_pcre_cache, 1);
 }
 
-#ifdef ZTS
 static void php_pcre_shutdown_globals(zend_pcre_globals *pcre_globals TSRMLS_DC)
 {
 	zend_hash_destroy(&pcre_globals->pcre_cache);
 }
-#endif
 
 
 static /* {{{ PHP_MINFO_FUNCTION(pcre) */
@@ -96,8 +101,11 @@ static PHP_MINIT_FUNCTION(pcre)
 	
 	REGISTER_LONG_CONSTANT("PREG_PATTERN_ORDER", PREG_PATTERN_ORDER, CONST_CS | CONST_PERSISTENT);
 	REGISTER_LONG_CONSTANT("PREG_SET_ORDER", PREG_SET_ORDER, CONST_CS | CONST_PERSISTENT);
+	REGISTER_LONG_CONSTANT("PREG_OFFSET_CAPTURE", PREG_OFFSET_CAPTURE, CONST_CS | CONST_PERSISTENT);
 	REGISTER_LONG_CONSTANT("PREG_SPLIT_NO_EMPTY", PREG_SPLIT_NO_EMPTY, CONST_CS | CONST_PERSISTENT);
 	REGISTER_LONG_CONSTANT("PREG_SPLIT_DELIM_CAPTURE", PREG_SPLIT_DELIM_CAPTURE, CONST_CS | CONST_PERSISTENT);
+	REGISTER_LONG_CONSTANT("PREG_SPLIT_OFFSET_CAPTURE", PREG_SPLIT_OFFSET_CAPTURE, CONST_CS | CONST_PERSISTENT);
+	REGISTER_LONG_CONSTANT("PREG_GREP_INVERT", PREG_GREP_INVERT, CONST_CS | CONST_PERSISTENT);
 	return SUCCESS;
 }
 /* }}} */
@@ -105,6 +113,10 @@ static PHP_MINIT_FUNCTION(pcre)
 /* {{{ PHP_MSHUTDOWN_FUNCTION(pcre) */
 static PHP_MSHUTDOWN_FUNCTION(pcre)
 {
+#ifndef ZTS
+	php_pcre_shutdown_globals(&pcre_globals TSRMLS_CC);
+#endif
+
 	return SUCCESS;
 }
 /* }}} */
@@ -121,7 +133,7 @@ static PHP_RINIT_FUNCTION(pcre)
 
 /* {{{ pcre_get_compiled_regex
  */
-static pcre* pcre_get_compiled_regex(char *regex, pcre_extra *extra, int *preg_options) {
+PHPAPI pcre* pcre_get_compiled_regex(char *regex, pcre_extra **extra, int *preg_options) {
 	pcre				*re = NULL;
 	int				 	 coptions = 0;
 	int				 	 soptions = 0;
@@ -150,7 +162,7 @@ static pcre* pcre_get_compiled_regex(char *regex, pcre_extra *extra, int *preg_o
 #if HAVE_SETLOCALE
 		if (!strcmp(pce->locale, locale)) {
 #endif
-			extra = pce->extra;
+			*extra = pce->extra;
 			*preg_options = pce->preg_options;
 			return pce->re;
 #if HAVE_SETLOCALE
@@ -272,7 +284,7 @@ static pcre* pcre_get_compiled_regex(char *regex, pcre_extra *extra, int *preg_o
 					  tables);
 
 	if (re == NULL) {
-		zend_error(E_WARNING, "Compilation failed: %s at offset %d\n", error, erroffset);
+		zend_error(E_WARNING, "Compilation failed: %s at offset %d", error, erroffset);
 		efree(pattern);
 		return NULL;
 	}
@@ -280,7 +292,7 @@ static pcre* pcre_get_compiled_regex(char *regex, pcre_extra *extra, int *preg_o
 	/* If study option was specified, study the pattern and
 	   store the result in extra for passing to pcre_exec. */
 	if (do_study) {
-		extra = pcre_study(re, soptions, &error);
+		*extra = pcre_study(re, soptions, &error);
 		if (error != NULL) {
 			zend_error(E_WARNING, "Error while studying pattern");
 		}
@@ -292,10 +304,10 @@ static pcre* pcre_get_compiled_regex(char *regex, pcre_extra *extra, int *preg_o
 
 	/* Store the compiled pattern and extra info in the cache. */
 	new_entry.re = re;
-	new_entry.extra = extra;
+	new_entry.extra = *extra;
 	new_entry.preg_options = poptions;
 #if HAVE_SETLOCALE
-	new_entry.locale = locale;
+	new_entry.locale = pestrdup(locale, 1);
 	new_entry.tables = tables;
 #endif
 	zend_hash_update(&PCRE_G(pcre_cache), regex, regex_len+1, (void *)&new_entry,
@@ -305,16 +317,40 @@ static pcre* pcre_get_compiled_regex(char *regex, pcre_extra *extra, int *preg_o
 }
 /* }}} */
 
+/* {{{ add_offset_pair */
+static inline void add_offset_pair(zval *result, char *str, int len, int offset, char *name)
+{
+	zval *match_pair;
+
+	ALLOC_ZVAL(match_pair);
+	array_init(match_pair);
+	INIT_PZVAL(match_pair);
+
+	/* Add (match, offset) to the return value */
+	add_next_index_stringl(match_pair, str, len, 1);
+	add_next_index_long(match_pair, offset);
+	
+	if (name) {
+		zval_add_ref(&match_pair);
+		zend_hash_update(Z_ARRVAL_P(result), name, strlen(name)+1, &match_pair, sizeof(zval *), NULL);
+	}
+	zend_hash_next_index_insert(Z_ARRVAL_P(result), &match_pair, sizeof(zval *), NULL);
+}
+/* }}} */
+
 /* {{{ php_pcre_match
  */
 static void php_pcre_match(INTERNAL_FUNCTION_PARAMETERS, int global)
 {
-	zval			**regex,			/* Regular expression */
-					**subject,			/* String to match against */
-					**subpats = NULL,	/* Array for subpatterns */
-					**subpats_order,	/* Order of the results in the subpatterns
-										   array for global match */
-					*result_set,		/* Holds a set of subpatterns after
+	/* parameters */
+	char  		    *regex;				/* Regular expression */
+	char 		    *subject;			/* String to match against */
+	int			     regex_len;
+	int				 subject_len;
+	zval 			*subpats = NULL;	/* Array for subpatterns */
+	int				 flags;				/* Match control flags */
+
+	zval			*result_set,		/* Holds a set of subpatterns after
 										   a global match */
 				   **match_sets = NULL;	/* An array of sets of matches for each
 										   subpattern after a global match */
@@ -326,61 +362,56 @@ static void php_pcre_match(INTERNAL_FUNCTION_PARAMETERS, int global)
 	int			 	*offsets;			/* Array of subpattern offsets */
 	int				 num_subpats;		/* Number of captured subpatterns */
 	int			 	 size_offsets;		/* Size of the offsets array */
-	int				 start_offset;		/* Where the new search starts */
+	int				 start_offset = 0;	/* Where the new search starts */
 	int			 	 matched;			/* Has anything matched */
-	int				 i;
-	int				 subpats_order_val = 0;	/* Integer value of subpats_order */
+	int				 subpats_order = 0; /* Order of subpattern matches */
+	int				 offset_capture = 0;/* Capture match offsets: yes/no */
 	int				 g_notempty = 0;	/* If the match should not be empty */
-	const char	   **stringlist;		/* Used to hold list of subpatterns */
+	const char	   **stringlist;		/* Holds list of subpatterns */
 	char			*match;				/* The current match */
+	char 		   **subpat_names = NULL;/* Array for named subpatterns */
+	int				 i;
 	
+	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, ((global) ? "ssz|ll" : "ss|zll"), &regex, &regex_len,
+							  &subject, &subject_len, &subpats, &flags, &start_offset) == FAILURE) {
+		RETURN_FALSE;
+	}
 	
-	/* Get function parameters and do error-checking. */
-	switch(ZEND_NUM_ARGS()) {
-		case 2:
-			if (global || zend_get_parameters_ex(2, &regex, &subject) == FAILURE) {
-				WRONG_PARAM_COUNT;
-			}
-			break;
-			
-		case 3:
-			if (zend_get_parameters_ex(3, &regex, &subject, &subpats) == FAILURE) {
-				WRONG_PARAM_COUNT;
-			}
-			if (global)
-				subpats_order_val = PREG_PATTERN_ORDER;
-			break;
+	if (global)
+		subpats_order = PREG_PATTERN_ORDER;
 
-		case 4:
-			if (zend_get_parameters_ex(4, &regex, &subject, &subpats, &subpats_order) == FAILURE) {
-				WRONG_PARAM_COUNT;
-			}
-	
-			/* Make sure subpats_order is a number */
-			convert_to_long_ex(subpats_order);
-			subpats_order_val = Z_LVAL_PP(subpats_order);
-			if (subpats_order_val < PREG_PATTERN_ORDER ||
-				subpats_order_val > PREG_SET_ORDER) {
-				zend_error(E_WARNING, "Wrong value for parameter 4 in call to preg_match_all()");
-			}
-			break;
-			
-		default:
-			WRONG_PARAM_COUNT;
+	if (ZEND_NUM_ARGS() > 3) {
+		offset_capture = flags & PREG_OFFSET_CAPTURE;
+		/*
+		 * subpats_order is pre-set to pattern mode so we change it only if
+		 * necessary.
+		 */
+		if (flags & 0xff) {
+			subpats_order = flags & 0xff;
+		}
+		if ((global && (subpats_order < PREG_PATTERN_ORDER || subpats_order > PREG_SET_ORDER)) ||
+			(!global && subpats_order != 0)) {
+			zend_error(E_WARNING, "Wrong value for parameter 4 in call to %s()", get_active_function_name(TSRMLS_C));
+			return;
+		}
 	}
 
-	/* Make sure we're dealing with strings. */
-	convert_to_string_ex(regex);
-	convert_to_string_ex(subject);
-
-	/* Make sure to clean up the passed array and initialize it. */
+	/* Overwrite the passed-in value for subpatterns with an empty array. */
 	if (subpats != NULL) {
-		zval_dtor(*subpats);
-		array_init(*subpats);
+		zval_dtor(subpats);
+		array_init(subpats);
+	}
+
+	/* Negative offset counts from the end of the string. */
+	if (start_offset < 0) {
+		start_offset = subject_len + start_offset;
+		if (start_offset < 0) {
+			start_offset = 0;
+		}
 	}
 
 	/* Compile regex or get it from cache. */
-	if ((re = pcre_get_compiled_regex(Z_STRVAL_PP(regex), extra, &preg_options)) == NULL) {
+	if ((re = pcre_get_compiled_regex(regex, &extra, &preg_options)) == NULL) {
 		RETURN_FALSE;
 	}
 
@@ -388,11 +419,36 @@ static void php_pcre_match(INTERNAL_FUNCTION_PARAMETERS, int global)
 	pcre_fullinfo(re, extra, PCRE_INFO_CAPTURECOUNT, &num_subpats);
 	num_subpats++;
 	size_offsets = num_subpats * 3;
-	offsets = (int *)emalloc(size_offsets * sizeof(int));
+	offsets = (int *)safe_emalloc(size_offsets, sizeof(int), 0);
 
-	/* Allocate match sets array and initialize the values */
-	if (global && subpats_order_val == PREG_PATTERN_ORDER) {
-		match_sets = (zval **)emalloc(num_subpats * sizeof(zval *));
+	/*
+	 * Build a mapping from subpattern numbers to their names. We will always
+	 * allocate the table, even though they may be no named subpatterns. This
+	 * avoids somewhat more complicated logic in the inner loops.
+	 */
+	subpat_names = (char **)safe_emalloc(num_subpats, sizeof(char *), 0);
+	memset(subpat_names, 0, sizeof(char *) * num_subpats);
+	{
+		int name_cnt, name_size, ni = 0;
+		char *name_table;
+		unsigned short name_idx;
+
+		pcre_fullinfo(re, extra, PCRE_INFO_NAMECOUNT, &name_cnt);
+		if (name_cnt > 0) {
+			pcre_fullinfo(re, extra, PCRE_INFO_NAMETABLE, &name_table);
+			pcre_fullinfo(re, extra, PCRE_INFO_NAMEENTRYSIZE, &name_size);
+
+			while (ni++ < name_cnt) {
+				name_idx = 0xff * name_table[0] + name_table[1];
+				subpat_names[name_idx] = name_table + 2;
+				name_table += name_size;
+			}
+		}
+	}
+
+	/* Allocate match sets array and initialize the values. */
+	if (global && subpats_order == PREG_PATTERN_ORDER) {
+		match_sets = (zval **)safe_emalloc(num_subpats, sizeof(zval *), 0);
 		for (i=0; i<num_subpats; i++) {
 			ALLOC_ZVAL(match_sets[i]);
 			array_init(match_sets[i]);
@@ -400,33 +456,30 @@ static void php_pcre_match(INTERNAL_FUNCTION_PARAMETERS, int global)
 		}
 	}
 
-	/* Start from the beginning of the string */
-	start_offset = 0;
 	match = NULL;
 	matched = 0;
 	
 	do {
 		/* Execute the regular expression. */
-		count = pcre_exec(re, extra, Z_STRVAL_PP(subject),
-						  Z_STRLEN_PP(subject), start_offset,
+		count = pcre_exec(re, extra, subject, subject_len, start_offset,
 						  exoptions|g_notempty, offsets, size_offsets);
 
 		/* Check for too many substrings condition. */	
 		if (count == 0) {
-			zend_error(E_NOTICE, "Matched, but too many substrings\n");
+			zend_error(E_NOTICE, "Matched, but too many substrings");
 			count = size_offsets/3;
 		}
 
 		/* If something has matched */
 		if (count >= 0) {
 			matched++;
-			match = Z_STRVAL_PP(subject) + offsets[0];
+			match = subject + offsets[0];
 
 			/* If subpatterns array has been passed, fill it in with values. */
 			if (subpats != NULL) {
 				/* Try to get the list of substrings and display a warning if failed. */
-				if (pcre_get_substring_list(Z_STRVAL_PP(subject),
-											offsets, count, &stringlist) < 0) {
+				if (pcre_get_substring_list(subject, offsets, count, &stringlist) < 0) {
+					efree(subpat_names);
 					efree(offsets);
 					efree(re);
 					zend_error(E_WARNING, "Get subpatterns list failed");
@@ -434,10 +487,16 @@ static void php_pcre_match(INTERNAL_FUNCTION_PARAMETERS, int global)
 				}
 
 				if (global) {	/* global pattern matching */
-					if (subpats_order_val == PREG_PATTERN_ORDER) {
+					if (subpats_order == PREG_PATTERN_ORDER) {
 						/* For each subpattern, insert it into the appropriate array. */
 						for (i = 0; i < count; i++) {
-							add_next_index_string(match_sets[i], (char *)stringlist[i], 1);
+							if (offset_capture) {
+								add_offset_pair(match_sets[i], (char *)stringlist[i],
+												offsets[(i<<1)+1] - offsets[i<<1], offsets[i<<1], NULL);
+							} else {
+								add_next_index_stringl(match_sets[i], (char *)stringlist[i],
+													   offsets[(i<<1)+1] - offsets[i<<1], 1);
+							}
 						}
 						/*
 						 * If the number of captured subpatterns on this run is
@@ -457,17 +516,36 @@ static void php_pcre_match(INTERNAL_FUNCTION_PARAMETERS, int global)
 						
 						/* Add all the subpatterns to it */
 						for (i = 0; i < count; i++) {
-							add_next_index_string(result_set, (char *)stringlist[i], 1);
+							if (offset_capture) {
+								add_offset_pair(result_set, (char *)stringlist[i],
+												offsets[(i<<1)+1] - offsets[i<<1], offsets[i<<1], subpat_names[i]);
+							} else {
+								if (subpat_names[i]) {
+									add_assoc_stringl(result_set, subpat_names[i], (char *)stringlist[i],
+														   offsets[(i<<1)+1] - offsets[i<<1], 1);
+								}
+								add_next_index_stringl(result_set, (char *)stringlist[i],
+													   offsets[(i<<1)+1] - offsets[i<<1], 1);
+							}
 						}
 						/* And add it to the output array */
-						zend_hash_next_index_insert(Z_ARRVAL_PP(subpats), &result_set,
-													sizeof(zval *), NULL);
+						zend_hash_next_index_insert(Z_ARRVAL_P(subpats), &result_set, sizeof(zval *), NULL);
 					}
-				}
-				else {			/* single pattern matching */
+				} else {			/* single pattern matching */
 					/* For each subpattern, insert it into the subpatterns array. */
 					for (i = 0; i < count; i++) {
-						add_next_index_string((*subpats), (char *)stringlist[i], 1);
+						if (offset_capture) {
+							add_offset_pair(subpats, (char *)stringlist[i],
+											offsets[(i<<1)+1] - offsets[i<<1],
+											offsets[i<<1], subpat_names[i]);
+						} else {
+							if (subpat_names[i]) {
+								add_assoc_stringl(subpats, subpat_names[i], (char *)stringlist[i],
+												  offsets[(i<<1)+1] - offsets[i<<1], 1);
+							}
+							add_next_index_stringl(subpats, (char *)stringlist[i],
+												   offsets[(i<<1)+1] - offsets[i<<1], 1);
+						}
 					}
 				}
 
@@ -479,7 +557,7 @@ static void php_pcre_match(INTERNAL_FUNCTION_PARAMETERS, int global)
 			   this is not necessarily the end. We need to advance
 			   the start offset, and continue. Fudge the offset values
 			   to achieve this, unless we're already at the end of the string. */
-			if (g_notempty != 0 && start_offset < Z_STRLEN_PP(subject)) {
+			if (g_notempty != 0 && start_offset < subject_len) {
 				offsets[0] = start_offset;
 				offsets[1] = start_offset + 1;
 			} else
@@ -497,20 +575,25 @@ static void php_pcre_match(INTERNAL_FUNCTION_PARAMETERS, int global)
 	} while (global);
 
 	/* Add the match sets to the output array and clean up */
-	if (global && subpats_order_val == PREG_PATTERN_ORDER) {
-		for (i=0; i<num_subpats; i++) {
-			zend_hash_next_index_insert(Z_ARRVAL_PP(subpats), &match_sets[i], sizeof(zval *), NULL);
+	if (global && subpats_order == PREG_PATTERN_ORDER) {
+		for (i = 0; i < num_subpats; i++) {
+			if (subpat_names[i]) {
+				zend_hash_update(Z_ARRVAL_P(subpats), subpat_names[i],
+								 strlen(subpat_names[i])+1, &match_sets[i], sizeof(zval *), NULL);
+			}
+			zend_hash_next_index_insert(Z_ARRVAL_P(subpats), &match_sets[i], sizeof(zval *), NULL);
 		}
 		efree(match_sets);
 	}
 	
 	efree(offsets);
+	efree(subpat_names);
 
 	RETVAL_LONG(matched);
 }
 /* }}} */
 
-/* {{{ proto int preg_match(string pattern, string subject [, array subpatterns])
+/* {{{ proto int preg_match(string pattern, string subject [, array subpatterns [, int flags [, int offset ]]])
    Perform a Perl-style regular expression match */
 PHP_FUNCTION(preg_match)
 {
@@ -518,7 +601,7 @@ PHP_FUNCTION(preg_match)
 }
 /* }}} */
 
-/* {{{ proto int preg_match_all(string pattern, string subject, array subpatterns [, int order])
+/* {{{ proto int preg_match_all(string pattern, string subject, array subpatterns [, int flags [, int offset]])
    Perform a Perl-style global regular expression match */
 PHP_FUNCTION(preg_match_all)
 {
@@ -528,16 +611,39 @@ PHP_FUNCTION(preg_match_all)
 
 /* {{{ preg_get_backref
  */
-static inline int preg_get_backref(const char *walk, int *backref)
+static int preg_get_backref(char **str, int *backref)
 {
-	if (*walk && *walk >= '0' && *walk <= '9')
+	register char in_brace = 0;
+	register char *walk = *str;
+
+	if (walk[1] == 0)
+		return 0;
+
+	if (*walk == '$' && walk[1] == '{') {
+		in_brace = 1;
+		walk++;
+	}
+	walk++;
+
+	if (*walk >= '0' && *walk <= '9') {
 		*backref = *walk - '0';
-	else
+		walk++;
+	} else
 		return 0;
 	
-	if (walk[1] && walk[1] >= '0' && walk[1] <= '9')
-		*backref = *backref * 10 + walk[1] - '0';
+	if (*walk && *walk >= '0' && *walk <= '9') {
+		*backref = *backref * 10 + *walk - '0';
+		walk++;
+	}
+
+	if (in_brace) {
+		if (*walk == 0 || *walk != '}')
+			return 0;
+		else
+			walk++;
+	}
 	
+	*str = walk;
 	return 1;	
 }
 /* }}} */
@@ -610,7 +716,7 @@ static int preg_do_eval(char *eval_str, int eval_str_len, char *subject,
 				continue;
 			}
 			segment = walk;
-			if (preg_get_backref(walk+1, &backref)) {
+			if (preg_get_backref(&walk, &backref)) {
 				if (backref < count) {
 					/* Find the corresponding string match and substitute it
 					   in instead of the backref */
@@ -629,8 +735,6 @@ static int preg_do_eval(char *eval_str, int eval_str_len, char *subject,
 				}
 				smart_str_appendl(&code, esc_match, esc_match_len);
 
-				/* Adjust the walk pointer */
-				walk += (backref > 9 ? 3 : 2);
 				segment = walk;
 
 				/* Clean up and reassign */
@@ -649,7 +753,7 @@ static int preg_do_eval(char *eval_str, int eval_str_len, char *subject,
 	/* Run the code */
 	if (zend_eval_string(code.c, &retval, compiled_string_description TSRMLS_CC) == FAILURE) {
 		efree(compiled_string_description);
-		zend_error(E_ERROR, "Failed evaluating code:\n%s\n", code);
+		zend_error(E_ERROR, "Failed evaluating code:\n%s", code.c);
 		/* zend_error() does not return in this case */
 	}
 	efree(compiled_string_description);
@@ -669,10 +773,10 @@ static int preg_do_eval(char *eval_str, int eval_str_len, char *subject,
 
 /* {{{ php_pcre_replace
  */
-char *php_pcre_replace(char *regex,   int regex_len,
-					   char *subject, int subject_len,
-					   zval *replace_val, int is_callable_replace,
-					   int *result_len, int limit TSRMLS_DC)
+PHPAPI char *php_pcre_replace(char *regex,   int regex_len,
+							  char *subject, int subject_len,
+							  zval *replace_val, int is_callable_replace,
+							  int *result_len, int limit TSRMLS_DC)
 {
 	pcre			*re = NULL;			/* Compiled regular expression */
 	pcre_extra		*extra = NULL;		/* Holds results of studying */
@@ -689,21 +793,21 @@ char *php_pcre_replace(char *regex,   int regex_len,
 	int				 backref;			/* Backreference number */
 	int				 eval;				/* If the replacement string should be eval'ed */
 	int				 start_offset;		/* Where the new search starts */
-	int				 g_notempty = 0;	/* If the match should not be empty */
-	int				 replace_len;		/* Length of replacement string */
+	int				 g_notempty=0;		/* If the match should not be empty */
+	int				 replace_len=0;		/* Length of replacement string */
 	char			*result,			/* Result of replacement */
-					*replace,			/* Replacement string */
+					*replace=NULL,		/* Replacement string */
 					*new_buf,			/* Temporary buffer for re-allocation */
 					*walkbuf,			/* Location of current replacement in the result */
 					*walk,				/* Used to walk the replacement string */
 					*match,				/* The current match */
 					*piece,				/* The current piece of subject */
-					*replace_end,		/* End of replacement string */
+					*replace_end=NULL,	/* End of replacement string */
 					*eval_result,		/* Result of eval or custom function */
 					 walk_last;			/* Last walked character */
 
 	/* Compile regex or get it from cache. */
-	if ((re = pcre_get_compiled_regex(regex, extra, &preg_options)) == NULL) {
+	if ((re = pcre_get_compiled_regex(regex, &extra, &preg_options)) == NULL) {
 		return NULL;
 	}
 
@@ -714,7 +818,6 @@ char *php_pcre_replace(char *regex,   int regex_len,
 			return NULL;
 		}
 	} else {
-		convert_to_string(replace_val);
 		replace = Z_STRVAL_P(replace_val);
 		replace_len = Z_STRLEN_P(replace_val);
 		replace_end = replace + replace_len;
@@ -722,10 +825,10 @@ char *php_pcre_replace(char *regex,   int regex_len,
 
 	/* Calculate the size of the offsets array, and allocate memory for it. */
 	size_offsets = (pcre_info(re, NULL, NULL) + 1) * 3;
-	offsets = (int *)emalloc(size_offsets * sizeof(int));
+	offsets = (int *)safe_emalloc(size_offsets, sizeof(int), 0);
 	
 	alloc_len = 2 * subject_len + 1;
-	result = emalloc(alloc_len * sizeof(char));
+	result = safe_emalloc(alloc_len, sizeof(char), 0);
 
 	/* Initialize */
 	match = NULL;
@@ -739,7 +842,7 @@ char *php_pcre_replace(char *regex,   int regex_len,
 		
 		/* Check for too many substrings condition. */
 		if (count == 0) {
-			zend_error(E_NOTICE, "Matched, but too many substrings\n");
+			zend_error(E_NOTICE, "Matched, but too many substrings");
 			count = size_offsets/3;
 		}
 
@@ -771,10 +874,9 @@ char *php_pcre_replace(char *regex,   int regex_len,
 							walk_last = 0;
 							continue;
 						}
-						if (preg_get_backref(walk+1, &backref)) {
+						if (preg_get_backref(&walk, &backref)) {
 							if (backref < count)
 								new_len += offsets[(backref<<1)+1] - offsets[backref<<1];
-							walk += (backref > 9) ? 3 : 2;
 							continue;
 						}
 					}
@@ -814,13 +916,12 @@ char *php_pcre_replace(char *regex,   int regex_len,
 							walk_last = 0;
 							continue;
 						}
-						if (preg_get_backref(walk+1, &backref)) {
+						if (preg_get_backref(&walk, &backref)) {
 							if (backref < count) {
 								match_len = offsets[(backref<<1)+1] - offsets[backref<<1];
 								memcpy(walkbuf, subject + offsets[backref<<1], match_len);
 								walkbuf += match_len;
 							}
-							walk += (backref > 9) ? 3 : 2;
 							continue;
 						}
 					}
@@ -849,7 +950,7 @@ char *php_pcre_replace(char *regex,   int regex_len,
 				new_len = *result_len + subject_len - start_offset;
 				if (new_len + 1 > alloc_len) {
 					alloc_len = new_len + 1; /* now we know exactly how long it is */
-					new_buf = emalloc(alloc_len * sizeof(char));
+					new_buf = safe_emalloc(alloc_len, sizeof(char), 0);
 					memcpy(new_buf, result, *result_len);
 					efree(result);
 					result = new_buf;
@@ -916,6 +1017,9 @@ static char *php_replace_in_subject(zval *regex, zval *replace, zval **subject,
 			if (Z_TYPE_P(replace) == IS_ARRAY && !is_callable_replace) {
 				/* Get current entry */
 				if (zend_hash_get_current_data(Z_ARRVAL_P(replace), (void **)&replace_entry) == SUCCESS) {
+					if (!is_callable_replace) {
+						convert_to_string_ex(replace_entry);
+					}
 					replace_value = *replace_entry;
 					zend_hash_move_forward(Z_ARRVAL_P(replace));
 				} else {
@@ -978,6 +1082,10 @@ static void preg_replace_impl(INTERNAL_FUNCTION_PARAMETERS, zend_bool is_callabl
 		zend_get_parameters_ex(ZEND_NUM_ARGS(), &regex, &replace, &subject, &limit) == FAILURE) {
 		WRONG_PARAM_COUNT;
 	}
+	if (!is_callable_replace && Z_TYPE_PP(replace) == IS_ARRAY && Z_TYPE_PP(regex) != IS_ARRAY) {
+		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Parameter mismatch, pattern is a string while replacement in an array.");
+		RETURN_FALSE;
+	}
 
 	SEPARATE_ZVAL(replace);
 	if (Z_TYPE_PP(replace) != IS_ARRAY)
@@ -1013,6 +1121,7 @@ static void preg_replace_impl(INTERNAL_FUNCTION_PARAMETERS, zend_bool is_callabl
 		/* For each subject entry, convert it to string, then perform replacement
 		   and add the result to the return_value array. */
 		while (zend_hash_get_current_data(Z_ARRVAL_PP(subject), (void **)&subject_entry) == SUCCESS) {
+			SEPARATE_ZVAL(subject_entry);
 			if ((result = php_replace_in_subject(*regex, *replace, subject_entry, &result_len, limit_val, is_callable_replace TSRMLS_CC)) != NULL) {
 				/* Add to return array */
 				switch(zend_hash_get_current_key(Z_ARRVAL_PP(subject), &string_key, &num_key, 0))
@@ -1072,8 +1181,10 @@ PHP_FUNCTION(preg_split)
 	int				 limit_val = -1;	/* Integer value of limit */
 	int				 no_empty = 0;		/* If NO_EMPTY flag is set */
 	int				 delim_capture = 0; /* If delimiters should be captured */
+	int				 offset_capture = 0;/* If offsets should be captured */
 	int				 count = 0;			/* Count of matched subpatterns */
 	int				 start_offset;		/* Where the new search starts */
+	int				 next_offset;		/* End of the last delimiter match + 1 */
 	int				 g_notempty = 0;	/* If the match should not be empty */
 	char			*match,				/* The current match */
 					*last_match;		/* Location of last match */
@@ -1094,6 +1205,7 @@ PHP_FUNCTION(preg_split)
 			convert_to_long_ex(flags);
 			no_empty = Z_LVAL_PP(flags) & PREG_SPLIT_NO_EMPTY;
 			delim_capture = Z_LVAL_PP(flags) & PREG_SPLIT_DELIM_CAPTURE;
+			offset_capture = Z_LVAL_PP(flags) & PREG_SPLIT_OFFSET_CAPTURE;
 		}
 	}
 	
@@ -1102,7 +1214,7 @@ PHP_FUNCTION(preg_split)
 	convert_to_string_ex(subject);
 	
 	/* Compile regex or get it from cache. */
-	if ((re = pcre_get_compiled_regex(Z_STRVAL_PP(regex), extra, &preg_options)) == NULL) {
+	if ((re = pcre_get_compiled_regex(Z_STRVAL_PP(regex), &extra, &preg_options)) == NULL) {
 		RETURN_FALSE;
 	}
 	
@@ -1111,10 +1223,11 @@ PHP_FUNCTION(preg_split)
 
 	/* Calculate the size of the offsets array, and allocate memory for it. */
 	size_offsets = (pcre_info(re, NULL, NULL) + 1) * 3;
-	offsets = (int *)emalloc(size_offsets * sizeof(int));
+	offsets = (int *)safe_emalloc(size_offsets, sizeof(int), 0);
 	
 	/* Start at the beginning of the string */
 	start_offset = 0;
+	next_offset = 0;
 	last_match = Z_STRVAL_PP(subject);
 	match = NULL;
 	
@@ -1126,7 +1239,7 @@ PHP_FUNCTION(preg_split)
 
 		/* Check for too many substrings condition. */
 		if (count == 0) {
-			zend_error(E_NOTICE, "Matched, but too many substrings\n");
+			zend_error(E_NOTICE, "Matched, but too many substrings");
 			count = size_offsets/3;
 		}
 				
@@ -1135,9 +1248,15 @@ PHP_FUNCTION(preg_split)
 			match = Z_STRVAL_PP(subject) + offsets[0];
 
 			if (!no_empty || &Z_STRVAL_PP(subject)[offsets[0]] != last_match) {
-				/* Add the piece to the return value */
-				add_next_index_stringl(return_value, last_match,
-									   &Z_STRVAL_PP(subject)[offsets[0]]-last_match, 1);
+
+				if (offset_capture) {
+					/* Add (match, offset) pair to the return value */
+					add_offset_pair(return_value, last_match, &Z_STRVAL_PP(subject)[offsets[0]]-last_match, next_offset, NULL);
+				} else {
+                	/* Add the piece to the return value */
+					add_next_index_stringl(return_value, last_match,
+								   	   &Z_STRVAL_PP(subject)[offsets[0]]-last_match, 1);
+				}
 
 				/* One less left to do */
 				if (limit_val != -1)
@@ -1145,15 +1264,22 @@ PHP_FUNCTION(preg_split)
 			}
 			
 			last_match = &Z_STRVAL_PP(subject)[offsets[1]];
+            next_offset = offsets[1];
 
 			if (delim_capture) {
 				int i, match_len;
 				for (i = 1; i < count; i++) {
 					match_len = offsets[(i<<1)+1] - offsets[i<<1];
-					if (!no_empty || match_len > 0)
-						add_next_index_stringl(return_value,
-											   &Z_STRVAL_PP(subject)[offsets[i<<1]],
-											   match_len, 1);
+					/* If we have matched a delimiter */
+					if (!no_empty || match_len > 0) {
+						if (offset_capture) {
+							add_offset_pair(return_value, &Z_STRVAL_PP(subject)[offsets[i<<1]], match_len, offsets[i<<1], NULL);
+						} else {
+							add_next_index_stringl(return_value,
+												   &Z_STRVAL_PP(subject)[offsets[i<<1]],
+												   match_len, 1);
+						}
+					}
 				}
 			}
 		} else { /* Failed to match */
@@ -1177,11 +1303,19 @@ PHP_FUNCTION(preg_split)
 		/* Advance to the position right after the last full match */
 		start_offset = offsets[1];
 	}
-	
+
+
 	if (!no_empty || start_offset != Z_STRLEN_PP(subject))
-		/* Add the last piece to the return value */
-		add_next_index_string(return_value,
-							  &Z_STRVAL_PP(subject)[start_offset], 1);
+	{
+		if (offset_capture) {
+			/* Add the last (match, offset) pair to the return value */
+			add_offset_pair(return_value, &Z_STRVAL_PP(subject)[start_offset], Z_STRLEN_PP(subject) - start_offset, start_offset, NULL);
+		} else {
+			/* Add the last piece to the return value */
+			add_next_index_stringl(return_value, last_match, Z_STRVAL_PP(subject) + Z_STRLEN_PP(subject) - last_match, 1);
+		}
+	}
+
 	
 	/* Clean up */
 	efree(offsets);
@@ -1229,7 +1363,7 @@ PHP_FUNCTION(preg_quote)
 	
 	/* Allocate enough memory so that even if each character
 	   is quoted, we won't run out of room */
-	out_str = emalloc(2 * Z_STRLEN_PP(in_str_arg) + 1);
+	out_str = safe_emalloc(2, Z_STRLEN_PP(in_str_arg), 1);
 	
 	/* Go through the string and quote necessary characters */
 	for(p = in_str, q = out_str; p != in_str_end; p++) {
@@ -1278,6 +1412,7 @@ PHP_FUNCTION(preg_grep)
 {
 	zval		   **regex,				/* Regular expression */
 				   **input,				/* Input array */
+				   **flags,
 			   	   **entry;				/* An entry in the input array */
 	pcre			*re = NULL;			/* Compiled regular expression */
 	pcre_extra		*extra = NULL;		/* Holds results of studying */
@@ -1287,15 +1422,18 @@ PHP_FUNCTION(preg_grep)
 	int			 	 count = 0;			/* Count of matched subpatterns */
 	char			*string_key;
 	ulong			 num_key;
+	zend_bool		 invert = 0;		/* Whether to return non-matching
+										   entries */
 	
 	/* Get arguments and do error checking */
 	
-	if (ZEND_NUM_ARGS() != 2 || zend_get_parameters_ex(ZEND_NUM_ARGS(), &regex, &input) == FAILURE) {
+	if (ZEND_NUM_ARGS() < 2 || ZEND_NUM_ARGS() > 3 ||
+		zend_get_parameters_ex(ZEND_NUM_ARGS(), &regex, &input, &flags) == FAILURE) {
 		WRONG_PARAM_COUNT;
 	}
 	
 	if (Z_TYPE_PP(input) != IS_ARRAY) {
-		zend_error(E_WARNING, "Secong argument to preg_grep() should be an array");
+		zend_error(E_WARNING, "Second argument to preg_grep() should be an array");
 		return;
 	}
 
@@ -1303,15 +1441,20 @@ PHP_FUNCTION(preg_grep)
 	
 	/* Make sure regex is a string */
 	convert_to_string_ex(regex);
+
+	if (ZEND_NUM_ARGS() > 2) {
+		convert_to_long_ex(flags);
+		invert = (Z_LVAL_PP(flags) & PREG_GREP_INVERT) ? 1 : 0;
+	}
 	
 	/* Compile regex or get it from cache. */
-	if ((re = pcre_get_compiled_regex(Z_STRVAL_PP(regex), extra, &preg_options)) == NULL) {
+	if ((re = pcre_get_compiled_regex(Z_STRVAL_PP(regex), &extra, &preg_options)) == NULL) {
 		RETURN_FALSE;
 	}
 
 	/* Calculate the size of the offsets array, and allocate memory for it. */
 	size_offsets = (pcre_info(re, NULL, NULL) + 1) * 3;
-	offsets = (int *)emalloc(size_offsets * sizeof(int));
+	offsets = (int *)safe_emalloc(size_offsets, sizeof(int), 0);
 	
 	/* Initialize return array */
 	array_init(return_value);
@@ -1329,12 +1472,13 @@ PHP_FUNCTION(preg_grep)
 
 		/* Check for too many substrings condition. */
 		if (count == 0) {
-			zend_error(E_NOTICE, "Matched, but too many substrings\n");
+			zend_error(E_NOTICE, "Matched, but too many substrings");
 			count = size_offsets/3;
 		}
 
-		/* If something matched */
-		if (count > 0) {
+		/* If the entry fits our requirements */
+		if ((count > 0 && !invert) ||
+			(count < 0 && invert)) {
 			(*entry)->refcount++;
 
 			/* Add to return array */
@@ -1377,13 +1521,13 @@ zend_module_entry pcre_module_entry = {
 	STANDARD_MODULE_HEADER,
    "pcre",
 	pcre_functions,
-		   PHP_MINIT(pcre),
-		   PHP_MSHUTDOWN(pcre),
-		   PHP_RINIT(pcre),
-		   NULL,
-		   PHP_MINFO(pcre),
+	PHP_MINIT(pcre),
+	PHP_MSHUTDOWN(pcre),
+	PHP_RINIT(pcre),
+	NULL,
+	PHP_MINFO(pcre),
 	NO_VERSION_YET,
-		   STANDARD_MODULE_PROPERTIES
+	STANDARD_MODULE_PROPERTIES
 };
 
 #ifdef COMPILE_DL_PCRE
@@ -1399,6 +1543,6 @@ ZEND_GET_MODULE(pcre)
  * tab-width: 4
  * c-basic-offset: 4
  * End:
- * vim600: sw=4 ts=4 tw=78 fdm=marker
- * vim<600: sw=4 ts=4 tw=78
+ * vim600: sw=4 ts=4 fdm=marker
+ * vim<600: sw=4 ts=4
  */
