@@ -2,7 +2,7 @@
    +----------------------------------------------------------------------+
    | PHP Version 4                                                        |
    +----------------------------------------------------------------------+
-   | Copyright (c) 1997-2002 The PHP Group                                |
+   | Copyright (c) 1997-2003 The PHP Group                                |
    +----------------------------------------------------------------------+
    | This source file is subject to version 2.02 of the PHP license,      |
    | that is bundled with this package in the file LICENSE, and is        |
@@ -16,7 +16,7 @@
    +----------------------------------------------------------------------+
  */
 
-/* $Id: mod_files.c,v 1.72.2.3 2002/09/04 13:50:38 kalowsky Exp $ */
+/* $Id: mod_files.c,v 1.83.2.7 2003/09/25 14:53:58 sniper Exp $ */
 
 #include "php.h"
 
@@ -39,6 +39,10 @@
 #include <fcntl.h>
 #include <errno.h>
 
+#if HAVE_UNISTD_H
+#include <unistd.h>
+#endif
+
 #include "php_session.h"
 #include "mod_files.h"
 #include "ext/standard/flock_compat.h"
@@ -58,6 +62,8 @@ ps_module ps_mod_files = {
 	PS_MOD(files)
 };
 
+/* If you change the logic here, please also update the error message in
+ * ps_files_open() appropriately */
 static int ps_files_valid_key(const char *key)
 {
 	size_t len;
@@ -118,6 +124,11 @@ static char *ps_files_path_create(char *buf, size_t buflen, ps_files *data, cons
 static void ps_files_close(ps_files *data)
 {
 	if (data->fd != -1) {
+#ifdef PHP_WIN32 
+		/* On Win32 locked files that are closed without being explicitly unlocked
+		   will be unlocked only when "system resources become available". */
+		flock(data->fd, LOCK_UN);
+#endif
 		close(data->fd);
 		data->fd = -1;
 	}
@@ -135,19 +146,29 @@ static void ps_files_open(ps_files *data, const char *key TSRMLS_DC)
 
 		ps_files_close(data);
 		
-		if (!ps_files_valid_key(key) || 
-				!ps_files_path_create(buf, sizeof(buf), data, key))
+		if (!ps_files_valid_key(key)) {
+			php_error_docref(NULL TSRMLS_CC, E_WARNING, "The session id contains invalid characters, valid characters are only a-z, A-Z and 0-9");
+			return;
+		}
+		if (!ps_files_path_create(buf, sizeof(buf), data, key))
 			return;
 		
 		data->lastkey = estrdup(key);
 		
 		data->fd = VCWD_OPEN_MODE(buf, O_CREAT | O_RDWR | O_BINARY, 0600);
-		if (data->fd != -1) 
+		
+		if (data->fd != -1) {
 			flock(data->fd, LOCK_EX);
 
-		if (data->fd == -1)
-			php_error(E_WARNING, "open(%s, O_RDWR) failed: %s (%d)", buf, 
+#ifdef F_SETFD
+			if (fcntl(data->fd, F_SETFD, 1)) {
+				php_error_docref(NULL TSRMLS_CC, E_WARNING, "fcntl(%d, F_SETFD, 1) failed: %s (%d)", data->fd, strerror(errno), errno);
+			}
+#endif
+		} else {
+			php_error_docref(NULL TSRMLS_CC, E_WARNING, "open(%s, O_RDWR) failed: %s (%d)", buf, 
 					strerror(errno), errno);
+		}
 	}
 }
 
@@ -164,7 +185,7 @@ static int ps_files_cleanup_dir(const char *dirname, int maxlifetime TSRMLS_DC)
 
 	dir = opendir(dirname);
 	if (!dir) {
-		php_error(E_NOTICE, "ps_files_cleanup_dir: opendir(%s) failed: %s (%d)\n", dirname, strerror(errno), errno);
+		php_error_docref(NULL TSRMLS_CC, E_NOTICE, "ps_files_cleanup_dir: opendir(%s) failed: %s (%d)\n", dirname, strerror(errno), errno);
 		return (0);
 	}
 
@@ -190,7 +211,7 @@ static int ps_files_cleanup_dir(const char *dirname, int maxlifetime TSRMLS_DC)
 				buf[dirname_len + entry_len + 1] = '\0';
 				/* check whether its last access was more than maxlifet ago */
 				if (VCWD_STAT(buf, &sbuf) == 0 && 
-						(now - sbuf.st_atime) > maxlifetime) {
+						(now - sbuf.st_mtime) > maxlifetime) {
 					VCWD_UNLINK(buf);
 					nrdels++;
 				}
@@ -215,7 +236,13 @@ PS_OPEN_FUNC(files)
 
 	data->fd = -1;
 	if ((p = strchr(save_path, ';'))) {
+		errno = 0;
 		data->dirdepth = (size_t) strtol(save_path, NULL, 10);
+		if (errno == ERANGE) {
+			efree(data);
+			PS_SET_MOD_DATA(0);
+			return FAILURE;
+		}
 		save_path = p + 1;
 	}
 	data->basedir_len = strlen(save_path);
@@ -228,6 +255,8 @@ PS_CLOSE_FUNC(files)
 {
 	PS_FILES_DATA;
 
+	if (!data) return FAILURE;
+	
 	ps_files_close(data);
 
 	if (data->lastkey) 
@@ -244,6 +273,8 @@ PS_READ_FUNC(files)
 	long n;
 	struct stat sbuf;
 	PS_FILES_DATA;
+	
+	if (!data) return FAILURE;
 
 	ps_files_open(data, key TSRMLS_CC);
 	if (data->fd < 0)
@@ -255,9 +286,18 @@ PS_READ_FUNC(files)
 	data->st_size = *vallen = sbuf.st_size;
 	*val = emalloc(sbuf.st_size);
 
+#if defined(HAVE_PREAD)
+	n = pread(data->fd, *val, sbuf.st_size, 0);
+#else
 	lseek(data->fd, 0, SEEK_SET);
 	n = read(data->fd, *val, sbuf.st_size);
+#endif
+
 	if (n != sbuf.st_size) {
+		if (n == -1)
+			php_error_docref(NULL TSRMLS_CC, E_WARNING, "read failed: %s (%d)", strerror(errno), errno);
+		else
+			php_error_docref(NULL TSRMLS_CC, E_WARNING, "read returned less bytes than requested");
 		efree(*val);
 		return FAILURE;
 	}
@@ -269,6 +309,8 @@ PS_WRITE_FUNC(files)
 {
 	long n;
 	PS_FILES_DATA;
+	
+	if (!data) return FAILURE;
 
 	ps_files_open(data, key TSRMLS_CC);
 	if (data->fd < 0)
@@ -282,11 +324,18 @@ PS_WRITE_FUNC(files)
 	if (vallen < (int)data->st_size)
 		ftruncate(data->fd, 0);
 
+#if defined(HAVE_PWRITE)
+	n = pwrite(data->fd, val, vallen, 0);
+#else
 	lseek(data->fd, 0, SEEK_SET);
 	n = write(data->fd, val, vallen);
+#endif
 
 	if (n != vallen) {
-		php_error(E_WARNING, "write failed: %s (%d)", strerror(errno), errno);
+		if (n == -1)
+			php_error_docref(NULL TSRMLS_CC, E_WARNING, "write failed: %s (%d)", strerror(errno), errno);
+		else
+			php_error_docref(NULL TSRMLS_CC, E_WARNING, "write wrote less bytes than requested");
 		return FAILURE;
 	}
 
@@ -297,6 +346,8 @@ PS_DESTROY_FUNC(files)
 {
 	char buf[MAXPATHLEN];
 	PS_FILES_DATA;
+	
+	if (!data) return FAILURE;
 
 	if (!ps_files_path_create(buf, sizeof(buf), data, key))
 		return FAILURE;
@@ -314,6 +365,8 @@ PS_GC_FUNC(files)
 {
 	PS_FILES_DATA;
 	
+	if (!data) return FAILURE;
+	
 	/* we don't perform any cleanup, if dirdepth is larger than 0.
 	   we return SUCCESS, since all cleanup should be handled by
 	   an external entity (i.e. find -ctime x | xargs rm) */
