@@ -1,22 +1,23 @@
 /* 
    +----------------------------------------------------------------------+
-   | PHP version 4.0                                                      |
+   | PHP Version 4                                                        |
    +----------------------------------------------------------------------+
-   | Copyright (c) 1997-2001 The PHP Group                                |
+   | Copyright (c) 1997-2003 The PHP Group                                |
    +----------------------------------------------------------------------+
-   | This source file is subject to version 2.02 of the PHP license,      |
+   | This source file is subject to version 3.0 of the PHP license,       |
    | that is bundled with this package in the file LICENSE, and is        |
-   | available at through the world-wide-web at                           |
-   | http://www.php.net/license/2_02.txt.                                 |
+   | available through the world-wide-web at the following url:           |
+   | http://www.php.net/license/3_0.txt.                                  |
    | If you did not receive a copy of the PHP license and are unable to   |
    | obtain it through the world-wide-web, please send a note to          |
    | license@php.net so we can mail you a copy immediately.               |
    +----------------------------------------------------------------------+
    | Authors: Zeev Suraski <zeev@zend.com>                                |
+   |          Timm Friebe <php_sybase_ct@thekid.de>                       |
    +----------------------------------------------------------------------+
 */
 
-/* $Id: php_sybase_ct.h,v 1.10 2001/02/26 06:07:24 andi Exp $ */
+/* $Id: php_sybase_ct.h,v 1.14.2.2 2003/08/29 14:22:13 sniper Exp $ */
 
 #ifndef PHP_SYBASE_CT_H
 #define PHP_SYBASE_CT_H
@@ -39,12 +40,14 @@ PHP_FUNCTION(sybase_pconnect);
 PHP_FUNCTION(sybase_close);
 PHP_FUNCTION(sybase_select_db);
 PHP_FUNCTION(sybase_query);
+PHP_FUNCTION(sybase_unbuffered_query);
 PHP_FUNCTION(sybase_free_result);
 PHP_FUNCTION(sybase_get_last_message);
 PHP_FUNCTION(sybase_num_rows);
 PHP_FUNCTION(sybase_num_fields);
 PHP_FUNCTION(sybase_fetch_row);
 PHP_FUNCTION(sybase_fetch_array);
+PHP_FUNCTION(sybase_fetch_assoc);
 PHP_FUNCTION(sybase_fetch_object);
 PHP_FUNCTION(sybase_data_seek);
 PHP_FUNCTION(sybase_result);
@@ -53,7 +56,8 @@ PHP_FUNCTION(sybase_field_seek);
 PHP_FUNCTION(sybase_min_client_severity);
 PHP_FUNCTION(sybase_min_server_severity);
 PHP_FUNCTION(sybase_fetch_field);
-
+PHP_FUNCTION(sybase_set_message_handler);
+PHP_FUNCTION(sybase_deadlock_retry_count);
 
 #include <ctpublic.h>
 
@@ -66,6 +70,8 @@ ZEND_BEGIN_MODULE_GLOBALS(sybase)
 	char *hostname;
 	char *server_message;
 	long min_server_severity, min_client_severity;
+	long deadlock_retry_count;
+	zval *callback_name;
 	CS_CONTEXT *context;
 ZEND_END_MODULE_GLOBALS(sybase)
 
@@ -75,6 +81,7 @@ typedef struct {
 	int valid;
 	int deadlock;
 	int dead;
+	int active_result_index;
 	long affected_rows;
 } sybase_link;
 
@@ -92,23 +99,23 @@ typedef struct {
 	sybase_link *sybase_ptr;
 	int cur_row,cur_field;
 	int num_rows,num_fields;
+	
+	/* For unbuffered reads */
+	CS_INT *lengths;
+	CS_SMALLINT *indicators;
+	char **tmp_buffer;
+	unsigned char *numerics;
+	CS_INT *types;
+	CS_DATAFMT *datafmt;
+	int blocks_initialized;
+	CS_RETCODE last_retcode;
+	int store;
 } sybase_result;
 
-
 #ifdef ZTS
-# define SybCtLS_D	zend_sybase_globals *sybase_globals
-# define SybCtLS_DC	, SybCtLS_D
-# define SybCtLS_C	sybase_globals
-# define SybCtLS_CC , SybCtLS_C
-# define SybCtG(v) (sybase_globals->v)
-# define SybCtLS_FETCH()	zend_sybase_globals *sybase_globals = ts_resource(sybase_globals_id)
+# define SybCtG(v) TSRMG(sybase_globals_id, zend_sybase_globals *, v)
 #else
-# define SybCtLS_D
-# define SybCtLS_DC
-# define SybCtLS_C
-# define SybCtLS_CC
 # define SybCtG(v) (sybase_globals.v)
-# define SybCtLS_FETCH()
 #endif
 
 #else
